"""Integrations template tags."""

from __future__ import unicode_literals

import warnings

from django import template

from djblets.deprecation import RemovedInDjblets20Warning


register = template.Library()


@register.simple_tag
def render_integration_config_status(integration, config):
    """Render the integration configuration's status.

    This is deprecated and no longer returns anything.

    Deprecated:
        1.0.11:
        This method no longer serves any purpose, due to major UI
        changes. It now returns an empty string.

    Args:
        integration (djblets.integrations.integrations.Integration, unused):
            The integration to which the config belongs.

        config (djblets.integrations.models.IntegrationConfig, unused):
            The configuration whose status is to be rendered.

    Returns:
        unicode:
        An empty string.
    """
    warnings.warn('{% render_integration_config_status %} is deprecated and '
                  'should no longer be called.',
                  RemovedInDjblets20Warning)

    return integration.render_config_status(config)



"""Middleware used for logging.

The middleware sets up some advanced logging capabilities for profiling
and exception logging.
"""

from __future__ import unicode_literals

import logging
import sys
import traceback

from django.conf import settings
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.db import connection
from django.http import Http404
from django.utils import six
from django.utils.six.moves import cStringIO as StringIO

try:
    # Django >= 1.7
    from django.db.backends import utils as db_backend_utils
    from django.db.backends.base.base import BaseDatabaseWrapper
except ImportError:
    # Django < 1.7
    from django.db.backends import (util as db_backend_utils,
                                    BaseDatabaseWrapper)

from djblets.log import init_logging, init_profile_logger, log_timed


logger = logging.getLogger(__name__)


class CursorDebugWrapper(db_backend_utils.CursorDebugWrapper):
    """Debug cursor for databases that stores tracebacks in query logs.

    This is a replacement for the default debug wrapper for database cursors
    which stores a traceback in :py:attr:`connection.queries
    <django.db.connection.queries>`.

    This will dramatically increase the overhead of having
    ``settings.DEBUG=True``, so use with caution.
    """

    def execute(self, *args, **kwargs):
        """Execute a SQL statement.

        This will record an entry in the query log and attach a stack trace.

        Args:
            *args (tuple):
                Positional arguments for the parent method.

            **kwargs (dict):
                Keyword arguments for the parent method.

        Returns:
            object:
            The result of the statement.
        """
        try:
            return super(CursorDebugWrapper, self).execute(*args, **kwargs)
        finally:
            self._store_traceback(traceback.format_stack())

    def executemany(self, *args, **kwargs):
        """Execute multiple SQL statements.

        This will record an entry in the query log and attach a stack trace.

        Note that the Django ORM does not appear to actually make use of this
        method. Third-parties can still use it.

        Args:
            *args (tuple):
                Positional arguments for the parent method.

            **kwargs (dict):
                Keyword arguments for the parent method.

        Returns:
            object:
            The result of the statement.
        """
        try:
            return super(CursorDebugWrapper, self).executemany(*args, **kwargs)
        finally:
            self._store_traceback(traceback.format_stack())

    def _store_traceback(self, stack):
        """Store a traceback in the query log.

        This will look up the correct variable for the query log and add the
        stack trace to the last entry.

        Args:
            stack (unicode):
                The formatted stack trace.
        """
        try:
            # Django >= 1.8
            queries = self.db.queries_log
        except AttributeError:
            # Django < 1.8
            queries = self.db.queries

        queries[-1]['stack'] = stack


# Return this cursor by default in the database backends. Custom backends
# potentially could override this.
BaseDatabaseWrapper.make_debug_cursor = \
    lambda self, cursor: CursorDebugWrapper(cursor, self)


def reformat_sql(sql):
    sql = sql.replace('`,`', '`, `')
    sql = sql.replace('SELECT ', 'SELECT\t')
    sql = sql.replace('` FROM ', '`\nFROM\t')
    sql = sql.replace(' WHERE ', '\nWHERE\t')
    sql = sql.replace(' INNER JOIN ', '\nINNER JOIN\t')
    sql = sql.replace(' LEFT OUTER JOIN ', '\nLEFT OUTER JOIN\t')
    sql = sql.replace(' OUTER JOIN ', '\nOUTER JOIN\t')
    sql = sql.replace(' ON ', '\n    ON ')
    sql = sql.replace(' ORDER BY ', '\nORDER BY\t')
    return sql


class LoggingMiddleware(object):
    """A piece of middleware that sets up page timing and profile logging.

    This is needed if using ``settings.LOGGING_PAGE_TIMES`` or
    ``settings.LOGGING_ALLOW_PROFILING``, in order to handle additional
    logging for page times and detailed profiling for debugging.
    """

    #: Exceptions that should be ignored by this logger.
    #:
    #: Each of these are handled by Django itself on the HTTP layer. We don't
    #: want to do any extra/unwanted logging of these.
    ignored_exceptions = (Http404, PermissionDenied, SuspiciousOperation)

    def process_request(self, request):
        """
        Processes an incoming request. This will set up logging.
        """
        if getattr(settings, 'LOGGING_PAGE_TIMES', False):
            request._page_timedloginfo = \
                log_timed('Page request: HTTP %s %s (by %s)' %
                          (request.method, request.path, request.user))

        if ('profiling' in request.GET and
            getattr(settings, "LOGGING_ALLOW_PROFILING", False)):
            settings.DEBUG = True

    def process_view(self, request, callback, callback_args, callback_kwargs):
        """
        Handler for processing a view. This will run the profiler on the view
        if profiling is allowed in the settings and the user specified the
        profiling parameter on the URL.
        """
        init_logging()

        if ('profiling' in request.GET and
            getattr(settings, "LOGGING_ALLOW_PROFILING", False)):
            import cProfile
            self.profiler = cProfile.Profile()
            args = (request,) + callback_args
            settings.DEBUG = True
            return self.profiler.runcall(callback, *args, **callback_kwargs)

    def process_response(self, request, response):
        """
        Handler for processing a response. Dumps the profiling information
        to the profile log file.
        """
        timedloginfo = getattr(request, '_page_timedloginfo', None)

        if timedloginfo:
            timedloginfo.done()

        if ('profiling' in request.GET and
            getattr(settings, "LOGGING_ALLOW_PROFILING", False)):

            init_profile_logger()

            self.profiler.create_stats()

            # Capture the stats
            out = StringIO()
            old_stdout, sys.stdout = sys.stdout, out
            self.profiler.print_stats(1)
            sys.stdout = old_stdout

            profile_log = logging.getLogger("profile")
            profile_log.log(logging.INFO,
                            "Profiling results for %s (HTTP %s):",
                            request.path, request.method)
            profile_log.log(logging.INFO, out.getvalue().strip())

            profile_log.log(logging.INFO,
                            '%d database queries made\n',
                            len(connection.queries))

            queries = {}
            for query in connection.queries:
                sql = reformat_sql(query['sql'])
                stack = ''.join(query['stack'][:-1])
                time = query['time']
                if sql in queries:
                    queries[sql].append((time, stack))
                else:
                    queries[sql] = [(time, stack)]

            times = {}
            for sql, entries in six.iteritems(queries):
                time = sum((float(entry[0]) for entry in entries))
                tracebacks = '\n\n'.join((entry[1] for entry in entries))
                times[time] = \
                    'SQL Query profile (%d times, %.3fs average)\n%s\n\n%s\n\n' % \
                    (len(entries), time / len(entries), sql, tracebacks)

            sorted_times = sorted(six.iterkeys(times), reverse=1)
            for time in sorted_times:
                profile_log.log(logging.INFO, times[time])

        return response

    def process_exception(self, request, exception):
        """Handle exceptions raised on a page.

        Logs the exception, along with the username and path where the
        exception occurred.

        Exceptions normally handled by Django's HTTP layer will be ignored.
        This helps avoid extraneous logging for 404 and Permission Denied
        errors.

        Args:
            request (django.http.HttpRequest):
                The HTTP request for the page.

            exception (Exception):
                The exception that was raised.
        """
        if not isinstance(exception, self.ignored_exceptions):
            logger.exception('Exception thrown for user %s at %s\n\n%s',
                             request.user, request.build_absolute_uri(),
                             exception,
                             request=request)



#
# __init__.py -- Basic version and package information
#
# Copyright (c) 2007-2013  Christian Hammond
# Copyright (c) 2007-2013  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import unicode_literals


# The version of Djblets
#
# This is in the format of:
#
#   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#
VERSION = (1, 0, 11, 'final', 0, True)


def get_version_string():
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%s" % VERSION[2]

    if VERSION[3] != 'final':
        if VERSION[3] == 'rc':
            version += ' RC%s' % VERSION[4]
        else:
            version += ' %s %s' % (VERSION[3], VERSION[4])

    if not is_release():
        version += " (dev)"

    return version


def get_package_version():
    version = '%d.%d' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%d" % VERSION[2]

    if VERSION[3] != 'final':
        version += '%s%d' % (VERSION[3], VERSION[4])

    return version


def is_release():
    return VERSION[5]


__version_info__ = VERSION[:-1]
__version__ = get_package_version()



"""Model and cache management for SiteConfiguration."""

from __future__ import unicode_literals

from django.contrib.sites.models import Site
from django.db import models
from django.utils import six

from djblets.siteconfig.signals import siteconfig_reloaded


_SITECONFIG_CACHE = {}


class SiteConfigurationManager(models.Manager):
    """Manages cached instances of a SiteConfiguration.

    This provides functions for retrieving the current
    :py:class:`~djblets.siteconfig.models.SiteConfiguration` instance and
    working with cache expiration. Consumers are expected to use
    :py:meth:`get_current` to retrieve their instance, and are also expected
    to use the :py:class:`~djblets.siteconfig.middleware.SettingsMiddleware`
    to manage expiration between server processes.
    """

    def get_current(self):
        """Return the site configuration for the active site.

        Multiple calls to this method for the same
        :py:class:`~django.contrib.sites.models.Site` will return the same
        instance, as long as the old instance has not expired. Callers should
        not store the result of this method, as it may not be valid for long.

        Returns:
            djblets.siteconfig.models.SiteConfiguration:
            The current site configuration for the active site.

        Raises:
            django.core.exceptions.ImproperlyConfigured:
                Site information wasn't configured in Django.
        """
        return self.get_for_site_id(Site.objects.get_current().pk)

    def get_for_site_id(self, site_id):
        """Return the site configuration for a specific site ID.

        Multiple calls to this method for the same
        :py:class:`~django.contrib.sites.models.Site` will return the same
        instance, as long as the old instance has not expired. Callers should
        not store the result of this method, as it may not be valid for long.

        Args:
            site (int):
                The ID of the site to retrieve the configuration for.

        Returns:
            djblets.siteconfig.models.SiteConfiguration:
            The current site configuration for the specified site.
        """
        try:
            siteconfig = _SITECONFIG_CACHE[site_id]
        except KeyError:
            siteconfig = self.model.objects.get(site_id=site_id)
            _SITECONFIG_CACHE[site_id] = siteconfig

        return siteconfig

    def clear_cache(self):
        """Clear the entire SiteConfiguration cache.

        The next call to :py:meth:`get_current` for any
        :py:class:`~django.contrib.sites.models.Site` will query the
        database.
        """
        global _SITECONFIG_CACHE

        _SITECONFIG_CACHE = {}

    def check_expired(self):
        """Check whether any SiteConfigurations have expired.

        If a :py:class:`~djblets.siteconfig.models.SiteConfiguration` has
        expired (another process/server has saved a more recent version),
        this method will expire the cache for the old version.

        If there are any listeners for the
        :py:data:`~djblets.siteconfig.signals.siteconfig_reloaded` signal,
        a new :py:class:`~djblets.siteconfig.models.SiteConfiguration`
        instance will be immediately loaded and the signal will fire.
        Otherwise, a new instance will not be loaded right away.

        This should be called on each HTTP request. It's recommended that
        consumers use
        :py:class:`~djblets.siteconfig.middleware.SettingsMiddleware` to do
        this. It can also be called manually for long-living processes that
        aren't bound to HTTP requests.

        .. versionchanged:: 1.0.3

           The :py:data:`~djblets.siteconfig.signals.siteconfig_reloaded`
           signal is now emitted with a newly-fetched instance if there are
           any listeners.
        """
        send_signal = siteconfig_reloaded.has_listeners()

        for site_id, siteconfig in six.iteritems(_SITECONFIG_CACHE.copy()):
            if siteconfig.is_expired():
                try:
                    # This is stale. Get rid of it so we can load it next time.
                    del _SITECONFIG_CACHE[site_id]
                except KeyError:
                    # Another thread probably took care of this. We're done
                    # with this one.
                    continue

                # If there are any listeners to the signal, then reload the
                # SiteConfiguration now and let consumers know about it.
                # If there aren't any listeners, we save a database query,
                # and the instance will be loaded next time a caller
                # requests it.
                if send_signal:
                    siteconfig_reloaded.send(
                        sender=None,
                        siteconfig=self.get_for_site_id(site_id),
                        old_siteconfig=siteconfig)



"""Tests for the utilities for rate-limiting login attempts."""

from __future__ import unicode_literals

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.test.utils import override_settings
from django.utils import six

from djblets.auth.ratelimit import is_ratelimited, Rate
from djblets.testing.testcases import TestCase


class RateLimitTests(TestCase):
    """Unit tests for djblets.auth.ratelimit."""

    def setUp(self):
        super(RateLimitTests, self).setUp()
        self.request_factory = RequestFactory()

    def tearDown(self):
        super(RateLimitTests, self).tearDown()
        cache.clear()

    def test_rate_parsing(self):
        """Testing Rate.parse"""
        test_rates = (
            ('100/s', Rate(100, 1)),
            ('100/10s', Rate(100, 10)),
            ('100/10', Rate(100, 10)),
            ('100/m', Rate(100, 60)),
            ('400/10m', Rate(400, 600)),
            ('1000/h', Rate(1000, 3600)),
            ('800/d', Rate(800, 24 * 60 * 60)),
        )

        for rate_str, rate in test_rates:
            self.assertEqual(rate, Rate.parse(rate_str))

    @override_settings(LOGIN_LIMIT_RATE='7/m')
    def test_unauthenticated_user(self):
        """Testing is_ratelimited with unauthenticated user"""
        request = self.request_factory.get('/')
        request.user = AnonymousUser()

        self.assertFalse(is_ratelimited(request, increment=False))

    @override_settings(LOGIN_LIMIT_RATE='blah')
    def test_invalid_rate_limit(self):
        """Testing is_ratelimited with invalid rate limit parameter"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        with self.assertRaises(ImproperlyConfigured) as context:
            is_ratelimited(request, increment=False)

        self.assertEqual(six.text_type(context.exception),
                         'LOGIN_LIMIT_RATE setting could not be parsed.')

    @override_settings(LOGIN_LIMIT_RATE='1/h')
    def test_rate_limit_exceeded(self):
        """Testing is_ratelimited when limit exceeded"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.assertFalse(is_ratelimited(request, increment=True))
        self.assertTrue(is_ratelimited(request, increment=True))



"""Functions for looking up DMARC entries in DNS."""

from __future__ import unicode_literals

import logging

import pkg_resources
from dns import resolver as dns_resolver
from publicsuffix import PublicSuffixList

from djblets.cache.backend import DEFAULT_EXPIRATION_TIME, cache_memoize


logger = logging.getLogger(__name__)


class DmarcPolicy(object):
    """Types of DMARC policies.

    These policies define what happens if an e-mail fails sender verification
    (such as if the :mailheader:`From` address is spoofed).
    """

    #: A policy is not set on the record.
    #:
    #: This can be interpreted as being the same as :py:attr:`NONE`.
    UNSET = 0

    #: No action should be taken if verification fails.
    #:
    #: E-mails will not be rejected or quarantined, and the DMARC signature
    #: can be ignored.
    NONE = 1

    #: E-mails should be quarantined/marked as spam if verification fails.
    QUARANTINE = 2

    #: E-mails should be rejected if verification fails.
    REJECT = 3

    _POLICY_TYPE_MAP = {
        'none': NONE,
        'quarantine': QUARANTINE,
        'reject': REJECT,
    }

    @classmethod
    def parse(cls, policy_str):
        """Return a policy type from a value in a DMARC record.

        Args:
            policy_str (unicode):
                The policy type from the record.

        Returns:
            int:
            One of :py:attr:`UNSET`, :py:attr:`NONE`, :py:attr:`QUARANTINE`,
            or :py:attr:`REJECT`.
        """
        return cls._POLICY_TYPE_MAP.get(policy_str.lower(), cls.UNSET)


class DmarcRecord(object):
    """Information on a DMARC record for a subdomain or organization domain.

    This is a parsed representation of the contents of a standard DMARC TXT
    record. It contains information that software can use to determine what
    will happen if a sender spoofs a From address, or if the e-mail otherwise
    fails sender verification.

    Senders can make use of this to determine whether they can safely spoof a
    :mailheader:`From` address (for legitimate reasons, such as to send an
    e-mail on behalf of a user when posting on a service), or whether they
    should fall back to an alternative means (such as using a noreply address
    and setting the :mailheader:`Reply-To` header).
    """

    def __init__(self, hostname, policy, subdomain_policy=DmarcPolicy.UNSET,
                 pct=100, fields={}):
        """Initialize the record.

        Args:
            hostname (unicode):
                The hostname containing the ``_dmarc.`` TXT record.

            policy (int):
                The sender policy defined for the record.

            subdomain_policy (int, optional):
                The sender policy defined for subdomains on this domain.

            pct (int, optional):
                The percentage (as a number from 0-100) of e-mails that should
                be subject to the sender policy.

            fields (dict, optional):
                Additional fields from the record.
        """
        self.hostname = hostname
        self.policy = policy
        self.subdomain_policy = subdomain_policy
        self.pct = pct
        self.fields = fields

    def __eq__(self, other):
        """Return whether two records are equal.

        Records are considered equal if they have the same
        :py:attr:`hostname` and :py:attr:`fields`.

        Args:
            other (DmarcRecord):
                The record to compare to.

        Returns:
            bool:
            ``True`` if the two records are equal. ``False`` if they are not.
        """
        return (self.hostname == other.hostname and
                self.fields == other.fields)

    @classmethod
    def parse(cls, hostname, txt_record):
        """Return a DmarcRecord from a DMARC TXT record.

        Args:
            hostname (unicode):
                The hostname owning the ``_dmarc.`` TXT record.

            txt_record (unicode):
                The TXT record contents representing the DMARC configuration.

        Returns:
            DmarcRecord:
            The parsed record, if this is a valid DMARC record. If this
            is not valid, ``None`` will be returned instead.
        """
        if not txt_record.startswith('"v=DMARC1'):
            # Records must start with this exact string. If this is missing,
            # it's not a valid DMARC record.
            return None

        fields = {}

        for part in txt_record.strip('"').split(';'):
            try:
                key, value = part.split('=', 1)
            except ValueError:
                # There's something wrong with this section of the record.
                # Or, it might be a trailing section after the final ";".
                # Ignore it.
                continue

            key = key.strip()
            value = value.strip()

            if key and value:
                fields[key] = value

        policy = DmarcPolicy.parse(fields.get('p', ''))
        subdomain_policy = DmarcPolicy.parse(fields.get('sp', ''))

        try:
            pct = int(fields.get('pct', '100'))
        except ValueError:
            pct = 100

        return cls(hostname=hostname,
                   policy=policy,
                   subdomain_policy=subdomain_policy,
                   pct=pct,
                   fields=fields)


def _fetch_dmarc_record(hostname, use_cache, cache_expiration):
    """Fetch a DMARC record from DNS, optionally caching it.

    This will query DNS for the DMARC record for a given hostname, returning
    the string contents of the record.

    The contents can be cached, preventing the need for subsequent DNS queries.

    This is used internally by :py:func:`get_dmarc_record`.

    Args:
        hostname (unicode):
            The hostname to fetch the record from.

        use_cache (bool, optional):
            Whether to use the cache for looking up and storing record data.

        cache_expiration (int, optional):
            The expiration time for cached data.

    Returns:
        DmarcRecord:
        The DMARC record. If it could not be found, or DNS lookup failed,
        ``None`` will be returned instead.
    """
    def _fetch_record():
        try:
            return dns_resolver.query('_dmarc.%s' % hostname,
                                      'TXT')[0].to_text()
        except (IndexError, dns_resolver.NXDOMAIN, dns_resolver.NoAnswer,
                dns_resolver.NoNameservers):
            raise ValueError

    try:
        if use_cache:
            record_str = cache_memoize(
                'dmarc-record-%s' % hostname,
                lambda: _fetch_record(),
                expiration=cache_expiration)
        else:
            record_str = _fetch_record()
    except ValueError:
        record_str = None

    if record_str:
        return DmarcRecord.parse(hostname, record_str)

    return None


def get_dmarc_record(hostname, use_cache=True,
                     cache_expiration=DEFAULT_EXPIRATION_TIME):
    """Return a DMARC record for a given hostname.

    This will query the DNS records for a hostname, returning a parsed version
    of the DMARC record, if found. If a record could not be found for the
    hostname, the organizational domain will be used instead (which is
    generally example.com for foo.bar.example.com, but this depends on the
    domain in question).

    By default, the fetched record from DNS is cached, allowing this to be
    called multiple times without repeated DNS queries. This is optional,
    as is the expiration time for the cached data (which defaults to 1 month).

    Args:
        hostname (unicode):
            The hostname to look up the DMARC information from.

        use_cache (bool, optional):
            Whether to use the cache for looking up and storing record data.

        cache_expiration (int, optional):
            The expiration time for cached data.

    Returns:
        DmarcRecord:
        The DMARC record. If it could not be found, ``None`` will be returned
        instead.
    """
    record = _fetch_dmarc_record(hostname=hostname,
                                 use_cache=use_cache,
                                 cache_expiration=cache_expiration)

    if not record:
        # We need to fetch from the Organizational Domain for the hostname
        # provided. For this, we need to look up from a Public Suffix list.
        # The third-party module 'publicsuffix' will help us find that
        # domain, in combination with a data file we must ship.
        filename = 'mail/public_suffix_list.dat'

        try:
            psl = PublicSuffixList(pkg_resources.resource_stream('djblets',
                                                                 filename))
        except IOError as e:
            logger.error('Unable to read public domain suffix list file '
                         '"%s" from Djblets package: %s',
                         filename, e)
        else:
            new_hostname = psl.get_public_suffix(hostname)

            if new_hostname != hostname:
                record = _fetch_dmarc_record(hostname=new_hostname,
                                             use_cache=use_cache,
                                             cache_expiration=cache_expiration)

    return record


def is_email_allowed_by_dmarc(email_address):
    """Return whether DMARC rules safely allow sending using an e-mail address.

    This will take an e-mail address (which must be in the form of
    ``name@domain``, ideally parsed by :py:func:`mail.utils.parseaddr`) and
    check to see if there are any DMARC rules that could prevent the e-mail
    from being sent/received if it were to fail sender verification.

    Callers can use this to decide whether they can safely send using a user's
    e-mail address, or whether they need to send using the service's address.

    Args:
        email_address (unicode):
            The e-mail address for the From field.

    Returns:
        bool:
        ``True`` if the e-mail address can be safely used in a
        :mailheader:`From` header. ``False`` if it should not be used.
    """
    hostname = email_address.split('@')[-1]
    bad_policies = (DmarcPolicy.QUARANTINE, DmarcPolicy.REJECT)

    dmarc_record = get_dmarc_record(hostname)

    return not (dmarc_record and
                dmarc_record.pct > 0 and
                ((dmarc_record.hostname == hostname and
                  dmarc_record.policy in bad_policies) or
                 (dmarc_record.hostname != hostname and
                  hostname.endswith(dmarc_record.hostname) and
                  dmarc_record.subdomain_policy in bad_policies)))



"""Specialized forms provided by Djblets.

This module provides a centralized place to load Djblets's utility forms.
From here, you can import:

* :py:class:`~djblets.forms.forms.key_value_form.KeyValueForm`
"""

from __future__ import unicode_literals

from djblets.forms.forms.key_value_form import KeyValueForm


__all__ = [
    'KeyValueForm',
]



"""Unit tests for djblets.configforms.views.ConfigPagesView."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.http import Http404
from django.test.client import RequestFactory
from django.utils import six

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.testing.testcases import TestCase


class TestForm1(ConfigPageForm):
    form_id = 'my-form-1'
    form_title = 'Form 1'

    def save(self):
        pass


class TestForm2(ConfigPageForm):
    form_id = 'my-form-2'
    form_title = 'Form 2'


class TestForm3(ConfigPageForm):
    form_id = 'my-form-3'
    form_title = 'Form 3'

    def is_visible(self):
        return False


class TestPage1(ConfigPage):
    page_id = 'my-page-1'
    form_classes = [TestForm1]


class TestPage2(ConfigPage):
    page_id = 'my-page-2'
    form_classes = [TestForm2]


class TestPage3(ConfigPage):
    page_id = 'my-page-3'
    form_classes = [TestForm3]


class MyConfigPagesView(ConfigPagesView):
    title = 'My Page Title'
    nav_title = 'My Nav Entry'
    page_classes = [TestPage1, TestPage2, TestPage3]

    css_bundle_names = ['my-css-bundle']
    js_bundle_names = ['my-js-bundle']

    js_model_class = 'MyModel'
    js_view_class = 'MyView'

    def get_js_model_data(self):
        return {
            'my-attr': 'value',
        }

    def get_js_view_data(self):
        return {
            'my-option': 'value',
        }


class ConfigPagesViewTests(TestCase):
    """Unit tests for djblets.configforms.views.ConfigPagesView."""

    def test_dispatch_initial_state(self):
        """Testing ConfigPagesView.dispatch initial state"""
        request = RequestFactory().request()
        request.user = User.objects.create(username='test-user')

        view = MyConfigPagesView()
        view.request = request

        response = view.dispatch(view.request)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(view.pages), 2)
        self.assertIsInstance(view.pages[0], TestPage1)
        self.assertIsInstance(view.pages[1], TestPage2)
        self.assertEqual(set(six.iterkeys(view.forms)),
                         {'my-form-1', 'my-form-2'})

    def test_get_context_data(self):
        """Testing ConfigPagesView.get_context_data"""
        request = RequestFactory().request()
        request.user = User.objects.create(username='test-user')

        view = MyConfigPagesView()
        view.request = request

        view.dispatch(view.request)

        self.assertEqual(
            view.get_context_data(),
            {
                'base_template_name': 'base.html',
                'page_title': 'My Page Title',
                'nav_title': 'My Nav Entry',
                'pages_id': 'config_pages',
                'pages': view.pages,
                'css_bundle_names': ['my-css-bundle'],
                'js_bundle_names': ['my-js-bundle'],
                'js_model_class': 'MyModel',
                'js_view_class': 'MyView',
                'js_model_data': {
                    'my-attr': 'value',
                },
                'js_view_data': {
                    'my-option': 'value',
                },
                'forms': list(six.itervalues(view.forms)),
                'render_sidebar': True,
            })

    def test_post_without_form_target(self):
        """Testing ConfigPagesView.dispatch with POST and no form_target"""
        request = RequestFactory().post('/config/')
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 400)

    def test_post_with_invalid_form_target(self):
        """Testing ConfigPagesView.dispatch with POST and invalid form_target
        """
        request = RequestFactory().post('/config/', {
            'form_target': 'bad',
        })
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()

        with self.assertRaises(Http404):
            view(request)

    def test_post_with_success(self):
        """Testing ConfigPagesView.dispatch with POST and success"""
        request = RequestFactory().post('/config/', {
            'form_target': 'my-form-1',
        })
        request.user = User.objects.create(username='test-user')
        request._dont_enforce_csrf_checks = True

        view = MyConfigPagesView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/config/')



"""Helpers for unit tests working with features."""

from __future__ import unicode_literals

from contextlib import contextmanager

from django.utils import six

from djblets.features.registry import get_features_registry


@contextmanager
def override_feature_checks(feature_states):
    """Override multiple features for a test.

    Unit tests can make use of this context manager to ensure that one or more
    features have particular enabled/disabled states before executing any code
    dependent on those features.

    Only the provided features will be modified, with all other feature logic
    falling back to the default behavior for the configured feature checker.

    Args:
        feature_states (dict):
            A dictionary of feature IDs to booleans (representing whether the
            feature is enabled).

    Example:
        .. code-block:: python

           feature_states = {
               'my-feature-1': True,
               'my-feature-2': False,
           }

           with override_feature_checks(feature_states):
               # Your test code here.
    """
    registry = get_features_registry()
    old_state = []

    for feature_id, enabled in six.iteritems(feature_states):
        feature = registry.get_feature(feature_id)

        old_state.append({
            'feature': feature,
            'func': feature.is_enabled,
        })

        feature.is_enabled = \
            lambda _is_enabled=enabled, **kwargs: _is_enabled

    yield

    for feature_info in old_state:
        feature_info['feature'].is_enabled = feature_info['func']


@contextmanager
def override_feature_check(feature_id, enabled):
    """Override a feature for a test.

    Unit tests can make use of this context manager to ensure that a specific
    feature has a particular enabled/disabled state before executing any code
    dependent on that feature.

    Only the provided feature will be modified, with all other feature logic
    falling back to the default behavior for the configured feature checker.

    Args:
        feature_id (unicode):
            The ID of the feature to override.

        enabled (bool):
            The enabled state for the feature.

    Example:
        .. code-block:: python

           with override_feature_check('my-feature', enabled=False):
               # Your test code here.
    """
    with override_feature_checks({feature_id: enabled}):
        yield



from __future__ import unicode_literals

import re
import uuid

from django.core.urlresolvers import ResolverMatch
from django.http import HttpRequest
from django.template import Context, Template

from djblets.extensions.extension import Extension, JSExtension
from djblets.extensions.hooks import TemplateHook
from djblets.extensions.tests import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


class TestJSExtension(JSExtension):
    model_class = 'FooNew'

    def get_model_data(self, request, **kwargs):
        return {'test': 'new'}


class TestJSExtensionDeprecated(JSExtension):
    model_class = 'FooOld'

    def get_model_data(self):
        return {'test': 'old'}


class TemplateTagTests(ExtensionTestsMixin, TestCase):
    """Tests for djblets.extensions.templatetags."""

    class TestExtension(Extension):
        css_bundles = {
            'default': {},
            'optional': {
                'apply_to': ['foo'],
            },
        }

        js_bundles = {
            'default': {},
            'optional': {
                'apply_to': ['foo'],
            },
        }

        js_extensions = [TestJSExtension, TestJSExtensionDeprecated]

    def setUp(self):
        super(TemplateTagTests, self).setUp()

        self.key = uuid.uuid4()

        self.extension = self.setup_extension(self.TestExtension,
                                              manager_key=self.key)
        self.request = HttpRequest()

    def test_template_hook_point_tag(self):
        """Testing template_hook_point template tag"""
        class DummyTemplateHook(TemplateHook):
            """A template hook where the content is specified."""

            def __init__(self, content, *args, **kwargs):
                super(DummyTemplateHook, self).__init__(*args, **kwargs)
                self._content = content

            def render_to_string(self, request, context):
                return self._content

        t = Template('{% load djblets_extensions %}'
                     '{% template_hook_point "test-hook" %}')

        DummyTemplateHook('foo\n', self.extension, 'test-hook')
        DummyTemplateHook('bar\n', self.extension, 'test-hook')

        self.assertEqual(
            t.render(Context({
                'request': self.request,
            })),
            'foo\nbar\n')

    def test_ext_static_tag(self):
        """Testing ext_static template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% ext_static ext "foo" %}')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '/ext/djblets.extensions.templatetags.tests.TestExtension/foo')

    def test_ext_css_bundle_tag(self):
        """Testing ext_css_bundle template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% ext_css_bundle ext "default" %}')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '/ext/djblets.extensions.templatetags.tests.TestExtension/css/'
            'default.min.css\n')

    def test_ext_js_bundle_tag(self):
        """Testing ext_js_bundle template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% ext_js_bundle ext "default" %}')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '/ext/djblets.extensions.templatetags.tests.TestExtension/js/'
            'default.min.js\n')

    def test_load_extensions_css_tag(self):
        """Testing load_extensions_css template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_css manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '/ext/djblets.extensions.templatetags.tests.TestExtension/css/'
            'default.min.css\n'
            '/ext/djblets.extensions.templatetags.tests.TestExtension/css/'
            'optional.min.css\n')

    def test_load_extensions_js_tag(self):
        """Testing load_extensions_js template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_js manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '/ext/djblets.extensions.templatetags.tests.TestExtension/js/'
            'default.min.js\n'
            '/ext/djblets.extensions.templatetags.tests.TestExtension/js/'
            'optional.min.js\n')

    def test_init_js_extensions(self):
        """Testing init_js_extensions template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% init_js_extensions manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        content = t.render(Context({
            'ext': self.extension,
            'manager_id': self.key,
            'request': self.request,
        }))

        self.assertIsNotNone(re.search(
            r'new FooNew\({\s+"test": "new",',
            content))

        self.assertIsNotNone(re.search(
            r'new FooOld\({\s+"test": "old",',
            content))



#
# djblets_deco.py -- Decorational template tags
#
# Copyright (c) 2007-2009  Christian Hammond
# Copyright (c) 2007-2009  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from django import template
from django.template.loader import render_to_string

from djblets.util.decorators import blocktag


register = template.Library()


@register.tag
@blocktag
def box(context, nodelist, classname=None):
    """
    Displays a box container around content, with an optional class name.
    """
    return render_to_string('deco/box.html', {
        'classname': classname or "",
        'content': nodelist.render(context)
    })


@register.tag
@blocktag
def errorbox(context, nodelist, box_id=None):
    """
    Displays an error box around content, with an optional ID.
    """
    return render_to_string('deco/errorbox.html', {
        'box_id': box_id or "",
        'content': nodelist.render(context)
    })



"""WebAPI scope dictionary tests."""

from __future__ import unicode_literals

from django.test.utils import override_settings
from django.utils import six
from oauth2_provider.settings import oauth2_settings

from djblets.extensions.extension import Extension
from djblets.extensions.manager import ExtensionManager
from djblets.extensions.models import RegisteredExtension
from djblets.extensions.testing.testcases import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase
from djblets.webapi.oauth2_scopes import (
    ExtensionEnabledWebAPIScopeDictionary,
    WebAPIScopeDictionary,
    enable_web_api_scopes,
    get_scope_dictionary)
from djblets.webapi.resources.mixins.oauth2_tokens import (
    ResourceOAuth2TokenMixin)
from djblets.webapi.testing.resources import make_resource_tree


class WebAPIScopeDictionaryTests(TestCase):
    """Tests for WebAPIScopeDictionary."""

    @classmethod
    def setUpClass(cls):
        super(WebAPIScopeDictionaryTests, cls).setUpClass()

        cls._resources = make_resource_tree(mixins=[
            ResourceOAuth2TokenMixin,
        ])

    def test_scope_dict(self):
        """Testing WebAPIScopeDictionary.scope_dict generates and caches scopes
        """
        scopes = WebAPIScopeDictionary(self._resources.root_resource)

        self.assertEqual(
            scopes.scope_dict,
            {
                'root:read': ('Ability to perform HTTP GET on the root '
                              'resource'),
                'item-child:read': ('Ability to perform HTTP GET on the '
                                    'item-child resource'),
                'item-child:write': ('Ability to perform HTTP PUT, POST on '
                                     'the item-child resource'),
                'list-child:read': ('Ability to perform HTTP GET on the '
                                    'list-child resource'),
                'list-child:destroy': ('Ability to perform HTTP DELETE on the '
                                       'list-child resource'),
                'parent:read': ('Ability to perform HTTP GET on the parent '
                                'resource'),
            })

    def test_getitem(self):
        """Testing WebAPIScopeDictionary.__getitem__"""
        scopes = WebAPIScopeDictionary(self._resources.root_resource)

        self.assertEqual(scopes['root:read'],
                         'Ability to perform HTTP GET on the root resource')

        with self.assertRaises(KeyError):
            scopes['bad-key']

    def test_contains(self):
        """Testing WebAPIScopeDictionary.__contains__"""
        scopes = WebAPIScopeDictionary(self._resources.root_resource)

        self.assertIn('root:read', scopes)
        self.assertNotIn('bad-key', scopes)

    def test_iterkeys(self):
        """Testing WebAPIScopeDictionary with six.iterkeys"""
        # Note that we have both implemented for Python 2/3 compatibility.
        # We'll actually be testing how this is used under the hood, with
        # six.iterkeys().
        scopes = WebAPIScopeDictionary(self._resources.root_resource)

        self.assertEqual(set(six.iterkeys(scopes)), {
            'root:read',
            'item-child:read',
            'item-child:write',
            'list-child:read',
            'list-child:destroy',
            'parent:read',
        })

    def test_clear(self):
        """Testing WebAPIScopeDictionary.clear"""
        scopes = WebAPIScopeDictionary(self._resources.root_resource)

        self.assertEqual(
            scopes.scope_dict,
            {
                'root:read': ('Ability to perform HTTP GET on the root '
                              'resource'),
                'item-child:read': ('Ability to perform HTTP GET on the '
                                    'item-child resource'),
                'item-child:write': ('Ability to perform HTTP PUT, POST on '
                                     'the item-child resource'),
                'list-child:read': ('Ability to perform HTTP GET on the '
                                    'list-child resource'),
                'list-child:destroy': ('Ability to perform HTTP DELETE on the '
                                       'list-child resource'),
                'parent:read': ('Ability to perform HTTP GET on the parent '
                                'resource'),
            })

        scopes.clear()
        self.assertEqual(scopes._scope_dict, {})


class ExtensionEnabledWebAPIScopeDictionaryTests(ExtensionTestCaseMixin,
                                                 TestCase):
    """Tests for ExtensionEnabledWebAPIScopeDictionary."""

    @classmethod
    def setUpClass(cls):
        super(ExtensionEnabledWebAPIScopeDictionaryTests, cls).setUpClass()

        cls._extension_manager = ExtensionManager('')

        cls._resources = make_resource_tree(mixins=[
            ResourceOAuth2TokenMixin,
        ], extension_manager=cls._extension_manager)

        class TestExtensionResource(cls._resources.base_resource):
            """An example resource on an extension."""

            name = 'test-ext'
            allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')

        cls._extension_resource = TestExtensionResource()

        class TestExtension(Extension):
            """An example extension."""

            extension_id = \
                'djblets.webapi.tests.test_oauth2_scopes.TestExtension'
            resources = [cls._extension_resource]
            registration = RegisteredExtension()

        cls.extension_class = TestExtension

    def test_scope_dict(self):
        """Testing ExtensionEnabledWebAPIScopeDictionary.scope_dict generates
        and caches scopes
        """
        self.extension_mgr.disable_extension(self.extension_class.extension_id)
        scopes = ExtensionEnabledWebAPIScopeDictionary(
            self._resources.root_resource)

        original_scope_dict = scopes.scope_dict

        base_scopes = {
            'root:read': 'Ability to perform HTTP GET on the root resource',
            'extension:read': ('Ability to perform HTTP GET on the '
                               'extension resource'),
            'extension:write': ('Ability to perform HTTP PUT on the '
                                'extension resource'),
            'item-child:read': ('Ability to perform HTTP GET on the '
                                'item-child resource'),
            'item-child:write': ('Ability to perform HTTP PUT, POST on '
                                 'the item-child resource'),
            'list-child:read': ('Ability to perform HTTP GET on the '
                                'list-child resource'),
            'list-child:destroy': ('Ability to perform HTTP DELETE on the '
                                   'list-child resource'),
            'parent:read': ('Ability to perform HTTP GET on the parent '
                            'resource'),
        }
        self.assertEqual(original_scope_dict, base_scopes)

        self.extension_mgr.enable_extension(self.extension_class.extension_id)
        self.assertEqual(scopes._scope_dict, {})

        new_base_scopes = dict({
            'test-ext:read': ('Ability to perform HTTP GET on the test-ext '
                              'resource'),
            'test-ext:write': ('Ability to perform HTTP PUT, POST on the '
                               'test-ext resource'),
            'test-ext:destroy': ('Ability to perform HTTP DELETE on the '
                                 'test-ext resource'),
        }, **base_scopes)

        new_scope_dict = scopes.scope_dict
        self.assertIs(new_scope_dict, original_scope_dict)
        self.assertEqual(new_scope_dict, new_base_scopes)

        self.extension_mgr.disable_extension(self.extension_class.extension_id)
        self.assertEqual(scopes._scope_dict, {})

        newest_scope_dict = scopes.scope_dict
        self.assertIs(newest_scope_dict, new_scope_dict)
        self.assertIs(newest_scope_dict, original_scope_dict)
        self.assertEqual(newest_scope_dict, base_scopes)


class ScopeEnablingTests(TestCase):
    """Tests for enabling WebAPI scopes at runtime."""

    @override_settings(
        WEB_API_ROOT_RESOURCE='djblets.webapi.tests.test_oauth2_auth.'
                              'root_resource',
    )
    def test_enable_webapi_scopes(self):
        """Testing enable_web_api_scopes()"""
        enable_web_api_scopes()
        scopes = get_scope_dictionary()

        self.assertIs(oauth2_settings.SCOPES, scopes)
        self.assertEqual(set(oauth2_settings._SCOPES), {
            'root:read',
            'item-child:read',
            'item-child:write',
            'list-child:read',
            'list-child:destroy',
            'parent:read',
        })



"""Internal support for handling deprecations in Djblets.

The version-specific objects in this module are not considered stable between
releases, and may be removed at any point. The base objects are considered
stable.
"""

from __future__ import unicode_literals

import warnings

from django.utils.functional import SimpleLazyObject


class BaseRemovedInDjbletsVersionWarning(DeprecationWarning):
    """Base class for a Djblets deprecation warning.

    All version-specific deprecation warnings inherit from this, allowing
    callers to check for Djblets deprecations without being tied to a specific
    version.
    """


class RemovedInDjblets20Warning(BaseRemovedInDjbletsVersionWarning):
    """Deprecations for features removed in Djblets 2.0.

    Note that this class will itself be removed in Djblets 2.0. If you need to
    check against Djblets deprecation warnings, please see
    :py:class:`BaseRemovedInDjbletsVersionWarning`. Alternatively, you can use
    the alias for this class, :py:data:`RemovedInNextDjbletsVersionWarning`.
    """


#: An alias for the next release of Djblets where features would be removed.
RemovedInNextDjbletsVersionWarning = RemovedInDjblets20Warning


def deprecated_arg_value(owner_name, value, old_arg_name, new_arg_name=None,
                         warning_cls=DeprecationWarning):
    """Wrap a value in a lazy object to warn when used.

    Args:
        owner_name (unicode):
            The name of the owner of this argument.

        value (object):
            The argument value.

        old_arg_name (unicode):
            The name of the argument that was deprecated.

        new_arg_name (unicode, optional):
            The optional name of the argument to use in the deprecated
            argument's place, if one is available.

        warning_cls (type, optional):
            The class to use for the warning. This should be
            :py:exc:`DeprecationWarning`, :py:exc:`PendingDeprecationWarning`,
            or a subclass of one.

    Returns:
        django.utils.functional.SimpleLazyObject:
        The value wrapped in a lazy object. The first time it is accessed,
        a warning will be emitted.
    """
    def _warn_on_use():
        if new_arg_name:
            warnings.warn('The "%s" argument for "%s" has been deprecated '
                          'and will be removed in a future version. Use "%s" '
                          'instead.'
                          % (old_arg_name, owner_name, new_arg_name),
                          warning_cls)
        else:
            warnings.warn('The "%s" argument for "%s" has been deprecated '
                          'and will be removed in a future version.'
                          % (old_arg_name, owner_name),
                          warning_cls)

        return value

    return SimpleLazyObject(_warn_on_use)



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.template import Context, Template

from djblets.testing.testcases import TagTest


class TagTests(TagTest):
    """Unit tests for gravatars template tags."""
    def test_gravatar_xss(self):
        """Testing {% gravatar %} doesn't allow XSS injection"""
        user = User(username='test',
                    first_name='"><script>alert(1);</script><"',
                    email='test@example.com')

        t = Template('{% load gravatars %}'
                     '{% gravatar user 32 %}')

        self.assertEqual(
            t.render(Context({
                'user': user,
            })),
            '<img src="https://secure.gravatar.com/avatar/'
            '55502f40dc8b7c769880b10874abc9d0?s=32" width="32" height="32" '
            'alt="&quot;&gt;&lt;script&gt;alert(1);&lt;/script&gt;&lt;&quot;" '
            'class="gravatar"/>')

    def test_gravatar_url_tag(self):
        """Testing gravatar_url template tag"""
        t = Template('{% load gravatars %}'
                     '{% gravatar_url "test@example.com" 32 %}')

        self.assertEqual(
            t.render(Context({})),
            'https://secure.gravatar.com/avatar/'
            '55502f40dc8b7c769880b10874abc9d0?s=32')



"""Tests for OAuth2 authentication with the web API."""

from __future__ import unicode_literals

from datetime import timedelta

from django.conf.urls import include, url
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User
from django.test.utils import override_settings
from django.utils import timezone
from oauth2_provider.models import AccessToken, Application

from djblets.webapi.auth.backends import reset_auth_backends
from djblets.webapi.auth.backends.oauth2_tokens import OAuth2TokenBackendMixin
from djblets.webapi.errors import (LOGIN_FAILED,
                                   NOT_LOGGED_IN,
                                   OAUTH_ACCESS_DENIED_ERROR,
                                   OAUTH_MISSING_SCOPE_ERROR)
from djblets.webapi.resources.mixins.oauth2_tokens import (
    ResourceOAuth2TokenMixin)
from djblets.webapi.testing.resources import make_resource_tree
from djblets.webapi.testing.testcases import TestCase, WebAPITestCaseMixin


resource_tree = make_resource_tree(
    mixins=[
        ResourceOAuth2TokenMixin,
    ],
    allow_anonymous_access=False,
)

root_resource = resource_tree.root_resource

urlpatterns = [
    url('^api/', include(root_resource.get_url_patterns())),
]


class OAuth2ModelBackend(OAuth2TokenBackendMixin, ModelBackend):
    """An OAuth2-enabled backend for unit tests."""


class OAuth2TokenAuthTests(WebAPITestCaseMixin, TestCase):
    """Tests for OAuth2 token authentication."""

    SETTINGS = {
        'AUTHENTICATION_BACKENDS': (
            'djblets.webapi.tests.test_oauth2_auth.OAuth2ModelBackend',
        ),
        'WEB_API_ROOT_RESOURCE': 'djblets.webapi.tests.test_oauth2_auth.'
                                 'root_resource',
        'WEB_API_AUTH_BACKENDS': (
            'djblets.webapi.auth.backends.oauth2_tokens.'
            'WebAPIOAuth2TokenAuthBackend',
        ),
        'ROOT_URLCONF': 'djblets.webapi.tests.test_oauth2_auth',
    }

    error_mimetype = 'application/json'

    @classmethod
    def setUpClass(cls):
        super(OAuth2TokenAuthTests, cls).setUpClass()

        reset_auth_backends()

    def tearDown(self):
        super(OAuth2TokenAuthTests, self).tearDown()

        reset_auth_backends()

    @override_settings(**SETTINGS)
    def setUp(self):
        super(OAuth2TokenAuthTests, self).setUp()

        self.owner = User.objects.create_user(username='app_owner',
                                              email='owner@example.com')

        self.application = Application.objects.create(
            name='Test Application',
            authorization_grant_type=Application.GRANT_IMPLICIT,
            redirect_uris='http://example.com/',
            user=self.owner,
        )

        self.user = User.objects.create_user(username='app_user',
                                             email='user@example.com')

    @override_settings(**SETTINGS)
    def test_get(self):
        """Testing WebAPI OAuth2 token access with valid scope"""
        token = self._create_token(scope='root:read')

        rsp = self.api_get(
            '/api/',
            expected_mimetype='application/json',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

    @override_settings(**SETTINGS)
    def test_get_unknown_scope(self):
        """Testing WebAPI OAuth2 token access with unknown scope"""
        token = self._create_token(scope='unknown')

        rsp = self.api_get(
            '/api/',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
            expected_status=403,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], OAUTH_MISSING_SCOPE_ERROR.code)

    @override_settings(**SETTINGS)
    def test_get_scope_insufficient(self):
        """Testing WebAPI OAuth2 token access with insufficient scope"""
        token = self._create_token(scope='')

        rsp = self.api_get(
            '/api/',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
            expected_status=403,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], OAUTH_MISSING_SCOPE_ERROR.code)

    @override_settings(**SETTINGS)
    def test_get_unallowed(self):
        """Testing WebAPI OAuth2 token access to a resource marked as
        inaccessible via OAuth2 tokens
        """
        token = self._create_token(scope='forbidden:read')

        rsp = self.api_get(
            '/api/forbidden/',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
            expected_status=403,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], OAUTH_ACCESS_DENIED_ERROR.code)

    @override_settings(**SETTINGS)
    def test_get_expired(self):
        """Testing WebAPI OAuth2 token access with an expired token
        """
        token = self._create_token(scope='root:read',
                                   expires=timedelta(hours=-1))

        rsp = self.api_get(
            '/api/',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
            expected_status=LOGIN_FAILED.http_status,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], LOGIN_FAILED.code)

    @override_settings(**SETTINGS)
    def test_get_session(self):
        """Testing WebAPI OAuth2 token access with a token stored in a session
        """
        token = self._create_token(scope='root:read')
        session_id = self._login_and_get_session(token)

        rsp = self.api_get(
            '/api/',
            HTTP_COOKIE='sessionid=%s' % session_id,
            expected_mimetype='application/json',
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

    @override_settings(**SETTINGS)
    def test_get_session_expired(self):
        """Testing WebAPI OAuth2 token access with an expired token stored in a
        session
        """
        token = self._create_token(scope='root:read')
        session_id = self._login_and_get_session(token)

        token.expires = timezone.now() - timedelta(hours=1)
        token.save(update_fields=('expires',))

        rsp = self.api_get(
            '/api/',
            HTTP_COOKIE='sessionid=%s' % session_id,
            expected_status=401,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], NOT_LOGGED_IN.code)

    @override_settings(**SETTINGS)
    def test_get_session_insufficient_scope(self):
        """Testing WebAPI OAuth2 token access with a token stored in a
        session with insufficient scope
        """
        token = self._create_token(scope='root:read')
        session_id = self._login_and_get_session(token)

        rsp = self.api_get(
            '/api/parents/',
            HTTP_COOKIE='sessionid=%s' % session_id,
            expected_status=403,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], OAUTH_MISSING_SCOPE_ERROR.code)

    @override_settings(**SETTINGS)
    def test_get_session_token_access_forbidden(self):
        """Testing WebAPI OAuth2 token access with a token stored in a
        session against a resource not allowing token access
        """
        token = self._create_token(scope='root:read forbidden:read')
        session_id = self._login_and_get_session(token)

        rsp = self.api_get(
            '/api/forbidden/',
            HTTP_COOKIE='sessionid=%s' % session_id,
            expected_status=403,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('err', rsp)
        err = rsp['err']
        self.assertIn('code', err)
        self.assertEqual(err['code'], OAUTH_ACCESS_DENIED_ERROR.code)

    def _login_and_get_session(self, token):
        """Log in with a token and return the session ID.

        Args:
            token (oauth2_provider.models.AccessToken):
                The access token to log in with.

        Returns:
            unicode:
            The session ID.
        """
        rsp, result = self.api_get_with_response(
            '/api/',
            HTTP_AUTHORIZATION='Bearer %s' % token.token,
            expected_mimetype='application/json',
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

        return result.cookies['sessionid'].value

    def _create_token(self, scope, expires=None):
        """Create an OAuth2 access token for testing.

        Args:
            scope (unicode):
                The scopes of the token.

            expires (datetime.timedelta, optional):
                How far into the future the token expires. If not provided,
                this argument defaults to one hour.

        Returns:
            oauth2_provider.models.AccessToken:
            The created access token.
        """
        if expires is None:
            expires = timedelta(hours=1)

        return AccessToken.objects.create(
            user=self.user,
            token='oauth-token',
            application=self.application,
            expires=timezone.now() + expires,
            scope=scope,
        )



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import six, timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from djblets.db.fields import JSONField
from djblets.webapi.managers import WebAPITokenManager
from djblets.webapi.signals import webapi_token_updated


@python_2_unicode_compatible
class BaseWebAPIToken(models.Model):
    """Base class for an access token used for authenticating with the API.

    Each token can be used to authenticate the token's owner with the API,
    without requiring a username or password to be provided. Tokens can
    be revoked, and new tokens added.

    Tokens can store policy information, which will later be used for
    restricting access to the API.
    """

    user = models.ForeignKey(User,
                             related_name='webapi_tokens',
                             on_delete=models.CASCADE)

    token = models.CharField(max_length=40, unique=True)
    time_added = models.DateTimeField(default=timezone.now)
    last_updated = models.DateTimeField(default=timezone.now)

    note = models.TextField(blank=True)
    policy = JSONField(null=True)

    extra_data = JSONField(null=True)

    objects = WebAPITokenManager()

    def is_accessible_by(self, user):
        return user.is_superuser or self.user == user

    def is_mutable_by(self, user):
        return user.is_superuser or self.user == user

    def is_deletable_by(self, user):
        return user.is_superuser or self.user == user

    def __str__(self):
        return 'Web API token for %s' % self.user

    def save(self, *args, **kwargs):
        """Save the token.

        If the token is being updated, the
        :py:data:`~djblets.webapi.signals.webapi_token_updated` signal will be
        emitted.

        Args:
            *args (tuple):
                Positional arguments to pass to the superclass.

            **kwargs (dict):
                Keyword arguments to pass to the superclass.
        """
        is_new = self.pk is None

        super(BaseWebAPIToken, self).save(*args, **kwargs)

        if not is_new:
            webapi_token_updated.send(instance=self, sender=type(self))

    @classmethod
    def get_root_resource(self):
        raise NotImplementedError

    @classmethod
    def validate_policy(cls, policy):
        """Validate an API policy.

        This will check to ensure that the policy is in a suitable format
        and contains the information required in a format that can be parsed.

        If a failure is found, a ValidationError will be raised describing
        the error and where it was found.
        """
        if not isinstance(policy, dict):
            raise ValidationError(_('The policy must be a JSON object.'))

        if not policy:
            # Empty policies are equivalent to allowing full access. If this
            # is empty, we can stop here.
            return

        if 'resources' not in policy:
            raise ValidationError(
                _('The policy is missing a "resources" section.'))

        resources_section = policy['resources']

        if not isinstance(resources_section, dict):
            raise ValidationError(
                _('The policy\'s "resources" section must be a JSON object.'))

        if not resources_section:
            raise ValidationError(
                _('The policy\'s "resources" section must not be empty.'))

        if '*' in resources_section:
            cls._validate_policy_section(resources_section, '*',
                                         'resources.*')

        resource_policies = [
            (section_name, section)
            for section_name, section in six.iteritems(resources_section)
            if section_name != '*'
        ]

        if resource_policies:
            valid_policy_ids = \
                cls._get_valid_policy_ids(cls.get_root_resource())

            for policy_id, section in resource_policies:
                if policy_id not in valid_policy_ids:
                    raise ValidationError(
                        _('"%s" is not a valid resource policy ID.')
                        % policy_id)

                for subsection_name, subsection in six.iteritems(section):
                    if not isinstance(subsection_name, six.text_type):
                        raise ValidationError(
                            _('%s must be a string in "resources.%s"')
                            % (subsection_name, policy_id))

                    cls._validate_policy_section(
                        section, subsection_name,
                        'resources.%s.%s' % (policy_id, subsection_name))

    @classmethod
    def _validate_policy_section(cls, parent_section, section_name,
                                 full_section_name):
        section = parent_section[section_name]

        if not isinstance(section, dict):
            raise ValidationError(
                _('The "%s" section must be a JSON object.')
                % full_section_name)

        if 'allow' not in section and 'block' not in section:
            raise ValidationError(
                _('The "%s" section must have "allow" and/or "block" '
                  'rules.')
                % full_section_name)

        if 'allow' in section and not isinstance(section['allow'], list):
            raise ValidationError(
                _('The "%s" section\'s "allow" rule must be a list.')
                % full_section_name)

        if 'block' in section and not isinstance(section['block'], list):
            raise ValidationError(
                _('The "%s" section\'s "block" rule must be a list.')
                % full_section_name)

    @classmethod
    def _get_valid_policy_ids(cls, resource, result=None):
        if result is None:
            result = set()

        if hasattr(resource, 'policy_id'):
            result.add(resource.policy_id)

        for child_resource in resource.list_child_resources:
            cls._get_valid_policy_ids(child_resource, result)

        for child_resource in resource.item_child_resources:
            cls._get_valid_policy_ids(child_resource, result)

        return result

    class Meta:
        abstract = True
        verbose_name = _('Web API token')
        verbose_name_plural = _('Web API tokens')



"""Base classes for implementing extensions."""

from __future__ import unicode_literals

import inspect
import locale
import logging
import os
import warnings
from email.parser import FeedParser

from django.conf import settings
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import get_mod_func
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext as _

from djblets.deprecation import RemovedInDjblets20Warning
from djblets.extensions.settings import Settings
from djblets.util.decorators import cached_property


logger = logging.getLogger(__name__)


class JSExtension(object):
    """Base class for a JavaScript extension.

    This can be subclassed to provide the information needed to initialize
    a JavaScript extension.

    The JSExtension subclass is expected to define a :py:attr:`model_class`
    attribute naming its JavaScript counterpart. This would be the variable
    name for the (uninitialized) model for the extension, defined in a
    JavaScript bundle.

    It may also define :py:attr:`apply_to`, which is a list of URL names that
    the extension will be initialized on. If not provided, the extension will
    be initialized on all pages.

    To provide additional data to the model instance, the JSExtension subclass
    can implement :py:meth:`get_model_data` and return a dictionary of data
    to pass. You may also override the :py:meth:`get_settings` method to
    return, a dict of settings to the :py:attr:`model_class`. By default, the
    associated extension's settings are returned.
    """

    #: The name of the JavaScript model class to instantiate.
    #:
    #: This class will be instantiated on the page. It should be a subclass of
    #: :js:class:`Djblets.Extension`.
    model_class = None

    #: The list of URL names to load this extension on.
    #:
    #: If not provided, this will be loaded on all pages.
    apply_to = None

    def __init__(self, extension):
        """Initialize the JavaScript extension.

        Args:
            extension (Extension):
                The main extension that owns this JavaScript extension.
        """
        self.extension = extension

    def applies_to(self, url_name):
        """Return whether this extension applies to the given URL name.

        Args:
            url_name (unicode):
                The name of the URL.

        Returns:
            bool:
            ``True`` if this JavaScript extension should load on the page
            with the given URL name. ``False`` if it should not load.
        """
        return self.apply_to is None or url_name in self.apply_to

    def get_model_data(self, request, **kwargs):
        """Return model data for the Extension model instance in JavaScript.

        Subclasses can override this to return custom data to pass to
        the extension class defined in :js:attr:`model_class`. This data must
        be JSON-serializable.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Model data to pass to the constructor of the JavaScript extension
            class.
        """
        return {}

    def get_settings(self):
        """Return the settings for the JS Extension.

        By default, this is the associated :py:class:`Extension` object's
        settings. Subclasses may override this method to provide different
        settings.

        These settings will be provided to the :py:attr:`model_class` as a
        ``settings`` key in its initialization options.

        Returns:
            dict:
            The extension settings.
        """
        return self.extension.settings


class Extension(object):
    """Base class for an extension.

    Extensions must subclass this class. They'll automatically have support for
    settings, adding hooks, and plugging into the administration UI.

    For information on writing extensions, see :ref:`writing-extensions`.

    Attributes:
        admin_site (django.contrib.admin.AdminSite):
            The database administration site set for the extension. This will
            be set automatically if :py:attr:`has_admin_site`` is ``True``.

        extension_manager (djblets.extensions.manager.ExtensionManager):
            The extension manager that manages this extension.

        hooks (set of djblets.extensions.hooks.ExtensionHook):
            The hooks currently registered and enabled for the extension.

        middleware_instances (list of object):
            The list of Django middleware instances. Each will be an instance
            of a class listed in :py:attr:`middleware`.

        settings (djblets.extensions.settings.Settings):
            The settings for the extension.
    """

    #: Metadata describing the package.
    #:
    #: This is used to set the name, version, and other information for the
    #: extension. This data defaults to coming from the extension's Python
    #: package metadata, but can be set on the extension itself. You would
    #: want to use this if shipping multiple extensions in a single package,
    #: if you want to include a space in the package name, or if you want
    #: the version to be set independently of the package, for example.
    #:
    #: The following keys are supported:
    #:
    #: * ``Name``: The displayed name of the extension.
    #: * ``Version``: The version of the extension.
    #: * ``Summary``: A summary of the extension.
    #: * ``Description``: A more detailed description of the extension.
    #: * ``License``: The license the extension was released under.
    #: * ``Author``: The name of the author writing/maintaining the extension.
    #: * ``Author-email``: The e-mail address of the author writing/maintaining
    #:   the extension.
    #: * ``Author-home-page``: The URL of the author writing/maintaining the
    #:   extension.
    #: * ``Home-page``: The URL of the extension's home/product/company page.
    metadata = None

    #: Whether or not the extension is user-configurable.
    #:
    #: If ``True``, the extension will have a :guilabel:`Configure` link in
    #: the extension list that will take the user to a page for modifying
    #: extension settings. The extension must provide a :file:`admin_urls.py`
    #: file defining a top-level URL (:regexp:`^$`) pointing to a view that
    #: will handle configuration.
    is_configurable = False

    #: Default settings for the extension.
    #:
    #: These values will be used when looking up keys in :py:attr:`settings`
    #: that don't have custom values saved in the database.
    default_settings = {}

    #: Whether or not the extension has a database administration site.
    #:
    #: If ``True``, the extension will have a :guilabel:`Database` link in
    #: the extension list that will take the user to a page for adding/editing
    #: any database models registered by the extension.
    has_admin_site = False

    #: A list of any extension IDs to enable when this extension is enabled.
    #:
    #: This is used to ensure that another extension is enabled before this
    #: one. It's primarily for extensions that are augmenting or depending on
    #: another extension's functionality.
    requirements = []

    #: A list of API resource instances offered by this extension.
    #:
    #: Each entry in the list is an instance of a custom
    #: :py:class:`~djblets.webapi.resources.WebAPIResource`. These resources
    #: will appear underneath the extension's own resource.
    resources = []

    #: A list of Django application module paths to load.
    #:
    #: Each of these will be added to :django:setting:`INSTALLED_APPS` while
    #: the extension is enabled. It follows the same format as that setting.
    apps = []

    #: A list of Django context processors to load.
    #:
    #: Each of these will be added to
    #: :django:setting:`TEMPLATE_CONTEXT_PROCESSORS` while the extension is
    #: enabled. It follows the same format as that setting.
    context_processors = []

    #: A list of Django middleware to load.
    #:
    #: Each of these will be added to
    #: :django:setting:`MIDDLEWARE_CLASSES` while the extension is enabled.
    #: It follows the same format as that setting.
    middleware = []

    #: A dictionary of CSS bundles to include in the package.
    #:
    #: These will be automatically packaged along with the extension, and can
    #: be loaded in templates using :py:func:`{% ext_css_bundle %}
    #: <djblets.extensions.templatetags.djblets_extensions.ext_css_bundle>`.
    #:
    #: Each key in the dictionary is the name of the bundle, and the
    #: value is a dictionary containing a ``source_filenames`` key pointing
    #: to a list of CSS/LessCSS files.
    #:
    #: A special bundle ID of ``default`` will cause the CSS to be loaded on
    #: every page.
    css_bundles = {}

    #: A dictionary of JavaScript bundles to include in the package.
    #:
    #: These will be automatically packaged along with the extension, and can
    #: be loaded in templates using :py:func:`{% ext_js_bundle %}
    #: <djblets.extensions.templatetags.djblets_extensions.ext_js_bundle>`.
    #:
    #: Each key in the dictionary is the name of the bundle, and the
    #: value is a dictionary containing a ``source_filenames`` key pointing
    #: to a list of JavaScript files.
    #:
    #: A special bundle ID of ``default`` will cause the JavaScript to be
    #: loaded on every page.
    js_bundles = {}

    #: A list of JavaScript extension classes to enable on pages.
    #:
    #: Each entry in the list is a :py:class:`JSExtension` subclass to load.
    js_extensions = []

    def __init__(self, extension_manager):
        """Initialize the extension.

        Subclasses should not override this. Instead, they should override
        :py:meth:`initialize`.

        Args:
            extension_manager (djblets.extensions.manager.ExtensionManager):
                The extension manager that manages this extension.
        """
        self.extension_manager = extension_manager
        self.hooks = set()
        self.settings = Settings(self)
        self.admin_site = None
        self.middleware_instances = []

        for middleware_cls in self.middleware:
            # We may be loading in traditional middleware (which doesn't take
            # any parameters in the constructor), or special Extension-aware
            # middleware (which takes an extension parameter). We need to
            # try to introspect and figure out what it is.
            try:
                arg_spec = inspect.getargspec(middleware_cls.__init__)
            except (AttributeError, TypeError):
                # There's no custom __init__ here. It may not exist
                # in the case of an old-style object, in which case we'll
                # get an AttributeError. Or, it may be a new-style object
                # with no custom __init__, in which case we'll get a TypeError.
                arg_spec = None

            if arg_spec and len(arg_spec) >= 2 and arg_spec[1] == 'extension':
                middleware_instance = middleware_cls(self)
            else:
                middleware_instance = middleware_cls()

            self.middleware_instances.append(middleware_instance)

        self.initialize()

    def initialize(self):
        """Initialize the extension.

        Subclasses can override this to provide any custom initialization.
        They do not need to call the parent function, as it does nothing.
        """
        pass

    def shutdown(self):
        """Shut down the extension.

        By default, this calls shutdown_hooks.

        Subclasses should override this if they need custom shutdown behavior.
        """
        self.shutdown_hooks()

    def shutdown_hooks(self):
        """Shut down all hooks for the extension."""
        for hook in self.hooks.copy():
            if hook.initialized:
                hook.disable_hook()

    def get_static_url(self, path):
        """Return the URL to a static media file for this extension.

        This takes care of resolving the static media path name to a path
        relative to the web server. If a versioned media file is found, it
        will be used, so that browser-side caching can be used.

        Args:
            path (unicode):
                The path within the static directory for the extension.

        Returns:
            unicode:
            The resulting static media URL.
        """
        return static('ext/%s/%s' % (self.id, path))

    def get_bundle_id(self, name):
        """Return the ID for a CSS or JavaScript bundle.

        This ID should be used when manually referencing the bundle in a
        template. The ID will be unique across all extensions.

        Args:
            name (unicode):
                The name of the bundle to reference.

        Returns:
            unicode:
            The ID of the bundle corresponding to the name.
        """
        return '%s-%s' % (self.id, name)

    @cached_property
    def admin_urlconf(self):
        """The module defining URLs for the extension's admin site."""
        try:
            name = '%s.%s' % (get_mod_func(self.__class__.__module__)[0],
                              'admin_urls')
            return __import__(name, {}, {}, [''])
        except Exception as e:
            raise ImproperlyConfigured(
                "Error while importing extension's admin URLconf %r: %s" %
                (name, e))


@python_2_unicode_compatible
class ExtensionInfo(object):
    """Information on an extension.

    This class stores the information and metadata on an extension. This
    includes the name, version, author information, where it can be downloaded,
    whether or not it's enabled or installed, and anything else that may be
    in the Python package for the extension.
    """

    encodings = ['utf-8', locale.getpreferredencoding(False), 'latin1']

    @classmethod
    def create_from_entrypoint(cls, entrypoint, ext_class):
        """Create a new ExtensionInfo from a Python EntryPoint.

        This will pull out information from the EntryPoint and return a new
        ExtensionInfo from it.

        It handles pulling out metadata from the older :file:`PKG-INFO` files
        and the newer :file:`METADATA` files.

        Args:
            entrypoint (pkg_resources.EntryPoint):
                The EntryPoint pointing to the extension class.

            ext_class (type):
                The extension class (subclass of :py:class:`Extension`).

        Returns:
            ExtensionInfo:
            An ExtensionInfo instance, populated with metadata from the
            package.
        """
        metadata = cls._get_metadata_from_entrypoint(entrypoint, ext_class.id)

        return cls(ext_class=ext_class,
                   package_name=metadata.get('Name'),
                   metadata=metadata)

    @classmethod
    def _get_metadata_from_entrypoint(cls, entrypoint, extension_id):
        """Return metadata information from an entrypoint.

        This is used internally to parse and validate package information from
        an entrypoint for use in ExtensionInfo.

        Args:
            entrypoint (pkg_resources.EntryPoint):
                The EntryPoint pointing to the extension class.

            extension_id (unicode):
                The extension's ID.

        Returns:
            dict:
            The resulting metadata dictionary.
        """
        dist = entrypoint.dist

        try:
            # Wheel, or other modern package.
            lines = dist.get_metadata_lines('METADATA')
        except IOError:
            try:
                # Egg, or other legacy package.
                lines = dist.get_metadata_lines('PKG-INFO')
            except IOError:
                lines = []
                logger.error('No METADATA or PKG-INFO found for the package '
                             'containing the %s extension. Information on '
                             'the extension may be missing.',
                             extension_id)

        # Both Python 2 and 3 will use byte strings here.
        data = b'\n'.join(lines)

        # Try to decode the PKG-INFO content. If no decoding method is
        # successful then the PKG-INFO content will remain unchanged and
        # processing will continue with the parsing.
        for enc in cls.encodings:
            try:
                data = data.decode(enc)
                break
            except UnicodeDecodeError:
                continue
        else:
            logger.warning('Failed decoding PKG-INFO content for extension %s',
                           entrypoint.name)

        p = FeedParser()
        p.feed(data)
        pkg_info = p.close()

        # Convert from a Message to a dictionary. Note that items() is correct
        # here. six.iteritems() will not work.
        return dict(pkg_info.items())

    def __init__(self, ext_class, package_name, metadata={}):
        """Instantiate the ExtensionInfo using metadata and an extension class.

        This will set information about the extension based on the metadata
        provided by the caller and the extension class itself.

        Args:
            ext_class (type):
                The extension class (subclass of :py:class:`Extension`).

            package_name (unicode):
                The package name owning the extension.

            metadata (dict, optional):
                Optional metadata for the extension. If the extension provides
                its own metadata, that will take precedence.

        Raises:
            TypeError:
                The parameters passed were invalid (they weren't a new-style
                call or a legacy entrypoint-related call).
        """
        try:
            issubclass(ext_class, Extension)
        except TypeError:
            try:
                is_entrypoint = (hasattr(ext_class, 'dist') and
                                 issubclass(package_name, Extension))
            except TypeError:
                is_entrypoint = False

            if is_entrypoint:
                # These are really (probably) an entrypoint and class,
                # respectively. Fix up the variables.
                entrypoint, ext_class = ext_class, package_name

                metadata = self._get_metadata_from_entrypoint(entrypoint,
                                                              ext_class.id)
                package_name = metadata.get('Name')

                # Warn after the above. Something about the above calls cause
                # warnings to be reset.
                warnings.warn(
                    'ExtensionInfo.__init__() no longer accepts an '
                    'EntryPoint. Please update your code to call '
                    'ExtensionInfo.create_from_entrypoint() instead.',
                    RemovedInDjblets20Warning)
            else:
                logger.error('Unexpected parameters passed to '
                             'ExtensionInfo.__init__: ext_class=%r, '
                             'package_name=%r, metadata=%r',
                             ext_class, package_name, metadata)

                raise TypeError(
                    _('Invalid parameters passed to ExtensionInfo.__init__'))

        # Set the base information from the extension and the package.
        self.package_name = package_name
        self.app_name = '.'.join(ext_class.__module__.split('.')[:-1])
        self.is_configurable = ext_class.is_configurable
        self.has_admin_site = ext_class.has_admin_site
        self.installed_htdocs_path = \
            os.path.join(settings.MEDIA_ROOT, 'ext', self.package_name)
        self.installed_static_path = \
            os.path.join(settings.STATIC_ROOT, 'ext', ext_class.id)

        # State set by ExtensionManager.
        self.enabled = False
        self.installed = False
        self.requirements = []

        # Set information from the provided metadata.
        if ext_class.metadata is not None:
            metadata.update(ext_class.metadata)

        self.metadata = metadata
        self.name = metadata.get('Name', package_name)
        self.version = metadata.get('Version')
        self.summary = metadata.get('Summary')
        self.description = metadata.get('Description')
        self.author = metadata.get('Author')
        self.author_email = metadata.get('Author-email')
        self.license = metadata.get('License')
        self.url = metadata.get('Home-page')
        self.author_url = metadata.get('Author-home-page', self.url)

    def __str__(self):
        return '%s %s (enabled = %s)' % (self.name, self.version, self.enabled)



"""Djblets registries.

Registries are collections that keep track of unique objects.

For information on writing registries, see
:ref:`the guide on writing registries <writing-registries>`.
"""

from __future__ import unicode_literals

import logging

from django.utils.translation import ugettext_lazy as _
from django.utils import six
from pkg_resources import iter_entry_points


from djblets.registries.errors import (AlreadyRegisteredError,
                                       ItemLookupError,
                                       RegistrationError)
from djblets.registries.signals import registry_populating


ALREADY_REGISTERED = 'already_registered'
ATTRIBUTE_REGISTERED = 'attribute_registered'
INVALID_ATTRIBUTE = 'invalid_attribute'
MISSING_ATTRIBUTE = 'missing_attribute'
UNREGISTER = 'unregister'
NOT_REGISTERED = 'not_registered'
LOAD_ENTRY_POINT = 'load_entry_point'


#: Default error messages for registries.
DEFAULT_ERRORS = {
    ALREADY_REGISTERED: _(
        'Could not register %(item)s: it is already registered.'
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register %(item)s: another item (%(duplicate)s) is already '
        'registered with %(attr_name)s = %(attr_value)s.'
    ),
    INVALID_ATTRIBUTE: _(
        '"%(attr_name)s" is not a registered lookup attribute.'
    ),
    LOAD_ENTRY_POINT: _(
        'Could not load entry point %(entry_point)s: %(error)s.',
    ),
    MISSING_ATTRIBUTE: _(
        'Could not register %(item)s: it does not have a "%(attr_name)s" '
        'attribute.'
    ),
    UNREGISTER: _(
        'Could not unregister %(item)s: it is not registered.'
    ),
    NOT_REGISTERED: _(
        'No item registered with %(attr_name)s = %(attr_value)s.'
    ),
}


class Registry(object):
    """An item registry.

    Item registries hold a set of objects that can be looked up by attributes.
    Each item is guaranteed to be unique and not share these attributes with
    any other item in the registry.
    """

    #: The name of the items being registered.
    item_name = None

    #: A list of attributes that items can be looked up by.
    lookup_attrs = []

    #: Error formatting strings for exceptions.
    #:
    #: Entries here override the global :py:data:`DEFAULT_ERRORS` dictionary
    #: for error messages.
    errors = {}

    #: The default error formatting strings.
    #:
    #: If subclasses need to provide additional errors that can be overridden,
    #: they should copy :py:data:`DEFAULT_ERRORS` and set their copy on the
    #: subclass as this attribute.
    default_errors = DEFAULT_ERRORS

    #: The error class indicating an already registered item.
    already_registered_error_class = AlreadyRegisteredError

    #: The lookup error exception class.
    lookup_error_class = ItemLookupError

    @property
    def populated(self):
        """Whether or not the registry is populated.

        Returns:
            bool: Whether or not the registry is populated.
        """
        return self._populated

    def __init__(self):
        """Initialize the registry."""
        self._registry = {}
        self._populated = False
        self._items = set()

        for attr_name in self.lookup_attrs:
            self._registry[attr_name] = {}

    def format_error(self, error_name, **error_kwargs):
        """Format an error message.

        Args:
            error_name (unicode):
                A symbolic name for the error, such as
                :py:data:`ALREADY_REGISTERED`.

            **error_kwargs (dict):
                The keyword arguments to provide to the error-specific
                formatting string.

        Returns:
            unicode:
                The formatted error message.
        """
        fmt = self.errors.get(error_name, self.default_errors.get(error_name))

        if fmt is None:
            raise ValueError('%s.format_error: Unknown error: "%s".',
                             type(self), error_name)

        return fmt % error_kwargs

    def get(self, attr_name, attr_value):
        """Return an item by its attribute value.

        Args:
            attr_name (unicode):
                The attribute name to look up an item by.

            attr_value (object):
                The corresponding attribute value.

        Returns:
            object: The registered item.

        Raises:
            djblets.registries.errors.ItemLookupError:
                When a lookup is attempted with an unsupported attribute, or
                the item cannot be found, this exception is raised.
        """
        self.populate()

        try:
            attr_map = self._registry[attr_name]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                INVALID_ATTRIBUTE, attr_name=attr_name))

        try:
            return attr_map[attr_value]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                NOT_REGISTERED, attr_name=attr_name, attr_value=attr_value))

    def register(self, item):
        """Register an item.

        Args:
            item (object):
                The item to register with the class.

        Raises:
            djblets.registries.errors.RegistrationError:
                Raised if the item is missing one of the required attributes.

            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the item is already registered or if the item shares
                an attribute name, attribute value pair with another item in
                the registry.
        """
        self.populate()
        attr_values = {}

        if item in self._items:
            raise self.already_registered_error_class(self.format_error(
                ALREADY_REGISTERED,
                item=item))

        for attr_name in self.lookup_attrs:
            attr_map = self._registry[attr_name]

            try:
                attr_value = getattr(item, attr_name)

                if attr_value in attr_map:
                    raise self.already_registered_error_class(
                        self.format_error(ATTRIBUTE_REGISTERED,
                                          item=item,
                                          duplicate=attr_map[attr_value],
                                          attr_name=attr_name,
                                          attr_value=attr_value))

                attr_values[attr_name] = attr_value
            except AttributeError:
                raise RegistrationError(self.format_error(
                    MISSING_ATTRIBUTE,
                    item=item,
                    attr_name=attr_name))

        for attr_name, attr_value in six.iteritems(attr_values):
            self._registry[attr_name][attr_value] = item

        self._items.add(item)

    def unregister_by_attr(self, attr_name, attr_value):
        """Unregister an item from the registry by an attribute.

        Args:
            attr_name (unicode):
                The name of the attribute.

            attr_value (object):
                The attribute value.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the attribute value is not found in the registry.
        """
        self.populate()

        try:
            attr_map = self._registry[attr_name]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                INVALID_ATTRIBUTE, attr_name=attr_name))
        try:
            item = attr_map[attr_value]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                NOT_REGISTERED, attr_name=attr_name, attr_value=attr_value))

        self.unregister(item)

    def unregister(self, item):
        """Unregister an item from the registry.

        Args:
            item (object):
                The item to unregister. This must be present in the registry.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the item is not found in the registry.
        """
        self.populate()

        try:
            self._items.remove(item)
        except KeyError:
            raise self.lookup_error_class(self.format_error(UNREGISTER,
                                                            item=item))

        for attr_name in self.lookup_attrs:
            attr_value = getattr(item, attr_name)
            del self._registry[attr_name][attr_value]

    def populate(self):
        """Ensure the registry is populated.

        Calling this method when the registry is populated will have no effect.
        """
        if self._populated:
            return

        self._populated = True

        for item in self.get_defaults():
            self.register(item)

        registry_populating.send(sender=type(self),
                                 registry=self)

    def get_defaults(self):
        """Return the default items for the registry.

        This method should be overridden by a subclass.

        Returns:
            list:
            The default items for the registry.
        """
        return []

    def reset(self):
        """Unregister all items and mark the registry unpopulated.

        This will result in the registry containing no entries. Any call to a
        method that would populate the registry will repopulate it.
        """
        if self._populated:
            for item in self._items.copy():
                self.unregister(item)

            self._populated = False

        assert len(self._items) == 0

    def __iter__(self):
        """Iterate through all items in the registry.

        This method does not provide a stable ordering.

        Yields:
            object: The items registered in this registry.
        """
        self.populate()

        for item in self._items:
            yield item

    def __len__(self):
        """Return the number of items in the registry.

        Returns:
            int: The number of items in the registry.
        """
        self.populate()
        return len(self._items)

    def __contains__(self, item):
        """Return whether or not the item is contained in the registry.

        Args:
            item (object):
                The item to look for.

        Returns:
            bool: Whether or not the item is contained in the registry.
        """
        self.populate()
        return item in self._items


class EntryPointRegistry(Registry):
    """A registry that auto-populates from an entry-point."""

    #: The entry point name.
    entry_point = None

    def get_defaults(self):
        """Yield the values from the entry point.

        Yields:
            object: The object from the entry point.
        """
        if self.entry_point is not None:
            entry_points = iter_entry_points(self.entry_point)

            for ep in entry_points:
                try:
                    yield self.process_value_from_entry_point(ep)
                except Exception as e:
                    logging.exception(self.format_error(LOAD_ENTRY_POINT,
                                                        entry_point=ep.name,
                                                        error=e))

    def process_value_from_entry_point(self, entry_point):
        """Return the item to register from the entry point.

        By default, this returns the loaded entry point.

        Args:
            entry_point (pkg_resources.EntryPoint):
                The entry point.

        Returns:
            object:
            The processed entry point value.
        """
        return entry_point.load()


class OrderedRegistry(Registry):
    """A registry that keeps track of registration order."""

    def __init__(self):
        """Initialize the OrderedRegistry"""
        super(OrderedRegistry, self).__init__()
        self._by_id = {}
        self._key_order = []

    def register(self, item):
        """Register an item.

        Args:
            item (object):
                The item to register with the class.

        Raises:
            djblets.registries.errors.RegistrationError:
                Raised if the item is missing one of the required attributes.

            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the item is already registered or if the item shares
                an attribute name, attribute value pair with another item in
                the registry.
        """
        super(OrderedRegistry, self).register(item)
        item_id = id(item)
        self._key_order.append(item_id)
        self._by_id[item_id] = item

    def unregister(self, item):
        """Unregister an item from the registry.

        Args:
            item (object):
                The item to unregister. This must be present in the registry.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the item is not found in the registry.
        """
        super(OrderedRegistry, self).unregister(item)
        item_id = id(item)
        del self._by_id[item_id]
        self._key_order.remove(item_id)

    def __iter__(self):
        """Yield the items in the order they were registered.

        Yields:
            object: The registered items.
        """
        self.populate()

        for key in self._key_order:
            yield self._by_id[key]

    def __getitem__(self, index):
        """Return an item by its registered index.

        Args:
            index (int):
                The position at which the item was registered. This is 0-based
                and negative indices are supported.

        Returns:
            object: The requested item.

        Raises:
            IndexError:
                This exception is raised if the requested index is out of
                range.

            TypeError:
                This exception is raised if the requested index is not an
                integer.
        """
        if not isinstance(index, int):
            raise TypeError('Index is not an integer (is %s).'
                            % type(index).__name__)

        # We don't have to call populate() because calling len() will.
        length = len(self)

        if index < 0:
            index += length

        if index > length:
            raise IndexError('Index is out of range.')

        return self._by_id[self._key_order[index]]



"""Deprecated module for authentication-related functions and classes.

.. deprecated:: 0.9

See the following for the new function/class locations:

* :py:func:`djblets.webapi.auth.backends.check_login`
* :py:func:`djblets.webapi.auth.backends.get_auth_backends`
* :py:func:`djblets.webapi.auth.backends.reset_auth_backends`
* :py:class:`djblets.webapi.auth.backends.base.WebAPIAuthBackend`
* :py:class:`djblets.webapi.auth.backends.basic.WebAPIBasicAuthBackend`
* :py:func:`djblets.webapi.auth.views.account_login`
* :py:func:`djblets.webapi.auth.views.account_logout`
"""

from __future__ import unicode_literals

from djblets.webapi.auth.backends import (check_login, get_auth_backends,
                                          reset_auth_backends)
from djblets.webapi.auth.backends.base import WebAPIAuthBackend
from djblets.webapi.auth.backends.basic import WebAPIBasicAuthBackend
from djblets.webapi.auth.views import account_login, account_logout


__all__ = [
    'WebAPIAuthBackend',
    'WebAPIBasicAuthBackend',
    'account_login',
    'account_logout',
    'check_login',
    'get_auth_backends',
    'reset_auth_backends',
]

__deprecated__ = __all__



"""Functions for safeguarding personally identifiable information."""

from __future__ import unicode_literals

import re

from django.conf import settings
from django.utils import six


#: A default list of unsafe keywords for URL patterns and querysets.
DEFAULT_PII_UNSAFE_URL_KEYWORDS = ('user', 'mail')


def build_pii_safe_page_url(url, url_kwargs=None, query_dict=None,
                            unsafe_keywords=None):
    """Return the current page URL with personal information redacted.

    This takes a URL and keywords used to match and build components of that
    URL and and looks for information to redact. It does this by inspecting
    ``url_kwargs`` and looking for any that contain certain keywords ("user"
    and "mail" by default) or could be e-mail addresses (containing a "@"),
    replacing them with "<REDACTED>". The redaction also applies to keys in the
    query string (if ``query_dict`` is provided).

    This can be used for services like Google Analytics, logging, or other
    purposes where a rough URL is needed that does not need to directly
    identify a specific location.

    Custom keywords can be set using the
    ``settings.DJBLETS_PII_UNSAFE_URL_KEYWORDS`` setting.

    Args:
        url (unicode):
            The URL to make safe.

        url_kwargs (dict, optional):
            A dictionary of keywords to values found in the URL. These keywords
            are matched against the list of unsafe keywords.

        query_dict (django.http.QueryDict, optional):
            An optional query dictionary representing a parsed querystring. If
            provided, the result will be appended to the URL.

        unsafe_keywords (set, optional):
            Custom keywords to match that are considered unsafe. This replaces
            the default keywords.

    Returns:
        unicode:
        The safe URL stripped of identifying information.
    """
    def _build_unsafe_keys(d):
        unsafe_keys = set()

        for key, value in six.iteritems(d):
            if isinstance(value, six.text_type):
                if '@' in value:
                    # This might be an e-mail address. Redact it.
                    unsafe_keys.add(key)
                else:
                    # Check for any keyword arguments that contain any unsafe
                    # keyword names as part of the key.
                    for unsafe_key in unsafe_keywords:
                        if unsafe_key in key:
                            unsafe_keys.add(key)
                            break

        return unsafe_keys

    unsafe_keywords = set(unsafe_keywords or
                          getattr(settings, 'DJBLETS_PII_UNSAFE_URL_KEYWORDS',
                                  DEFAULT_PII_UNSAFE_URL_KEYWORDS))

    cur_unsafe_url_keys = _build_unsafe_keys(url_kwargs or {})
    cur_unsafe_qs_keys = _build_unsafe_keys(query_dict or {})

    if cur_unsafe_url_keys:
        # This is not perfect. It's entirely possible that part of what we're
        # trying to replace would appear elsewhere in the URL, and we'd redact
        # more than what we intended.
        #
        # Ideally, we'd grab the original regex and try to fill out the URL
        # based on the resolved match and redacted data, but we don't have
        # that information handy.
        #
        # Really, this is just for the purposes of providing URLs in places
        # that need some informative URL without PII, so this will suffice.
        new_url = re.sub(
            '(%s)' % '|'.join(
                re.escape(url_kwargs[key])
                for key in cur_unsafe_url_keys
            ),
            '<REDACTED>',
            url)
    else:
        new_url = url

    if cur_unsafe_qs_keys:
        new_query = query_dict.copy()

        # We want to set this through a loop instead of using update() in
        # order to properly replace keys instead of appending to them when
        # using a QueryDict.
        for key in cur_unsafe_qs_keys:
            new_query[key] = '<REDACTED>'
    else:
        new_query = query_dict

    if new_query:
        new_url = '%s?%s' % (new_url, new_query.urlencode(safe='/<>'))

    return new_url


def build_pii_safe_page_url_for_request(request, unsafe_keywords=None):
    """Return the current page URL with personal information redacted.

    This wraps :py:func:`build_pii_safe_page_url`, returning a PII-safe URL
    based on the URL pattern used for the current page.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        unsafe_keywords (set, optional):
            Custom keywords to match that are considered unsafe. This replaces
            the default keywords.

    Returns:
        unicode:
        The safe URL stripped of identifying information.
    """
    if request.resolver_match is None:
        url_kwargs = {}
    else:
        url_kwargs = request.resolver_match.kwargs

    return build_pii_safe_page_url(url=request.path,
                                   url_kwargs=url_kwargs,
                                   query_dict=request.GET,
                                   unsafe_keywords=unsafe_keywords)



"""An authentication backend for HTTP Basic Auth."""

from __future__ import unicode_literals

import logging

from djblets.webapi.auth.backends.base import WebAPIAuthBackend


logger = logging.getLogger(__name__)


class WebAPIBasicAuthBackend(WebAPIAuthBackend):
    """Handles HTTP Basic Authentication for the web API."""

    www_auth_scheme = 'Basic realm="Web API"'

    def get_credentials(self, request):
        parts = request.META['HTTP_AUTHORIZATION'].split(' ')
        realm = parts[0]

        if realm != 'Basic':
            return None

        try:
            encoded_auth = parts[1]
            username, password = encoded_auth.decode('base64').split(':', 1)
        except Exception:
            logger.warning('Failed to parse HTTP_AUTHORIZATION header %s',
                           request.META['HTTP_AUTHORIZATION'],
                           exc_info=True,
                           extra={'request': request})
            return None

        return {
            'username': username,
            'password': password,
        }



from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.db.models.query_utils import Q
from django.utils import six

from djblets.db.query import chainable_select_related_queryset, prefix_q
from djblets.testing.testcases import TestCase


class ChainableSelectRelatedQuerySetTests(TestCase):
    """Tests for djblets.db.query.chainable_select_related_queryset."""

    def test_select_related_with_new_fields(self):
        """Testing chainable_select_related_queryset with
        select_related(field, ...)
        """
        queryset = User.objects.select_related('foo')
        queryset = chainable_select_related_queryset(queryset)
        queryset = queryset.select_related('foo__bar', 'baz')
        queryset = queryset.select_related('foobar', 'baz')

        self.assertEqual(
            queryset.query.select_related,
            {
                'foo': {
                    'bar': {},
                },
                'baz': {},
                'foobar': {},
            })

    def test_select_related_with_new_no_args(self):
        """Testing chainable_select_related_queryset with select_related()"""
        queryset = User.objects.select_related('foo')
        queryset = chainable_select_related_queryset(queryset)
        queryset = queryset.select_related()

        self.assertTrue(queryset.query.select_related)

    def test_select_related_with_none(self):
        """Testing chainable_select_related_queryset with select_related(None)
        """
        queryset = User.objects.select_related('foo')
        queryset = chainable_select_related_queryset(queryset)
        queryset = queryset.select_related(None)

        self.assertFalse(queryset.query.select_related)


class PrefixTests(TestCase):
    """Tests for djblets.db.query.prefix_q."""

    def test_simple(self):
        """Testing prefix_q prefixes simple expressions"""
        self.assertEqual(
            six.text_type(prefix_q('fk', Q(hello='goodbye'))),
            six.text_type(Q(fk__hello='goodbye')))

    def test_nested(self):
        """Testing prefix_q prefixes nested expressions"""
        self.assertEqual(
            six.text_type(prefix_q(
                'fk',
                Q(foo='foo') |
                (Q(bar='bar') &
                 Q(baz='baz')))),
            six.text_type(
                Q(fk__foo='foo') |
                (Q(fk__bar='bar') &
                 Q(fk__baz='baz'))))

    def test_bytestring_result(self):
        """Testing that prefix_q generates byte strings for key names"""
        q = prefix_q('fk', Q(foo='bar'))
        self.assertEqual(len(q.children), 1)
        self.assertIs(type(q.children[0]), tuple)
        self.assertIsInstance(q.children[0][0], six.binary_type)



"""Database models for privacy-related data storage."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from djblets.db.fields import JSONField


@python_2_unicode_compatible
class StoredConsentData(models.Model):
    """Stored information about a user's current and past consent decisions.

    This tracks what parts of a product a user has consented to, and how those
    decisions have changed in the past. The model is not meant to be queried
    directly by applications, but rather used behind the scenes by
    :py:class:`~djblets.privacy.consent.tracker.DatabaseConsentTracker`.

    Entries are associated with a user in the database (when available), but
    are also associated with a piece of identifying data for the purposes of
    privacy auditing. The identifying data (stored in
    :py:attr:`audit_identifier`) can be erased at any point without impacting
    the model, and can also store encrypted values if needed by the
    application.
    """

    user = models.OneToOneField(
        User,
        null=True,
        on_delete=models.SET_NULL,
        help_text=_('The user whose consent decisions are being stored.'))

    audit_identifier = models.CharField(
        max_length=255,
        db_index=True,
        blank=True,
        null=True,
        help_text=_('An identifier for locating the audit record as needed. '
                    'The value must be able to be compared against '
                    'information from the original user, but may be erased '
                    'or encrypted at any point.'))

    time_added = models.DateTimeField(
        auto_now_add=True,
        help_text=_('The date and time when the stored consent data was first '
                    'added to the database.'))

    last_updated = models.DateTimeField(
        auto_now=True,
        help_text=_('The date and time when the stored consent data was last '
                    'updated.'))

    #: The current consent grant state of each consent-tracking item.
    #:
    #: This is in the form of::
    #:
    #:     {
    #:         "<consent_requirement_id>": bool,
    #:         ...
    #:     }
    #:
    #: Each boolean indicates if consent is granted or denied.
    consent_grants = JSONField(
        default=dict,
        help_text=_('The consents granted or denied by the user.'))

    #: An audit trail of each consent-tracking decision.
    #:
    #: This is in the form of::
    #:
    #:     {
    #:         "<consent_requirement_id>": [
    #:             {
    #:                 "identifier": string,
    #:                 "granted": bool,
    #:                 "timestamp": UTC datetime,
    #:                 "source": string,
    #:                 "memo": string,
    #:             },
    #:             ...
    #:         ],
    #:         ...
    #:     }
    audit_trail = JSONField(
        default=dict,
        help_text=_('A record of all consent changes made by the user.'))

    def __str__(self):
        """Return a string representation of the object.

        Returns:
            unicode:
            The string representation of the object.
        """
        return 'Stored support data for %s' % (self.user or
                                               self.audit_identifier)

    class Meta:
        # Ensure we're using the same name across all versions of Django.
        db_table = 'djblets_privacy_storedconsentdata'

        ordering = ('-last_updated',)

        verbose_name = _('Stored consent data')
        verbose_name_plural = _('Stored consent data')



"""Mixins for test cases that need to test enabled extensions."""

from __future__ import unicode_literals

from djblets.extensions.extension import ExtensionInfo
from djblets.extensions.manager import get_extension_managers
from djblets.extensions.models import RegisteredExtension


class ExtensionTestCaseMixin(object):
    """Unit tests mixin for more easily testing extensions.

    This will do the hard work of creating the fake registration information
    needed for an extension and to instantiate an instance for testing.

    Subclasses need to define :py:attr:`extension_class` and may want to
    implement :py:meth:`get_extension_manager` (by default, the first
    registered extension manager will be used).

    Projects may want to provide their own subclass for their extensions to use
    that implements :py:meth:`get_extension_manager`, so extensions won't have
    to.

    Attributes:
        extension_mgr (djblets.extensions.manager.ExtensionManager):
            The extension manager owning the extension. Tests can use this to
            manually enable/disable the extension, if needed.

        extension (djblets.extensions.extension.Extension):
            The extension instance being tested.
    """

    #: The extension class to test.
    extension_class = None

    #: Optional metadata to use for the extension information.
    extension_metadata = {}

    #: Optional package name to use for the extension information.
    extension_package_name = 'TestPackage'

    def setUp(self):
        super(ExtensionTestCaseMixin, self).setUp()

        self.extension_mgr = self.get_extension_manager()

        # We want to override all the information, even if a previous test
        # already set it. The metadata may be different, and the registration
        # definitely needs to be replaced (as it contains extension settings).
        extension_id = '%s.%s' % (self.extension_class.__module__,
                                  self.extension_class.__name__)

        self.extension_class.id = extension_id
        self.extension_class.info = ExtensionInfo(
            ext_class=self.extension_class,
            package_name=self.extension_package_name,
            metadata=self.extension_metadata)

        self.extension_class.registration = RegisteredExtension.objects.create(
            class_name=extension_id,
            name=self.extension_class.info.name,
            enabled=True,
            installed=True)

        # We're going to manually inject the extension, instead of calling
        # load(), since it might not be found otherwise.
        self.extension_mgr._extension_classes[extension_id] = \
            self.extension_class

        self.extension = self.extension_mgr.enable_extension(extension_id)
        assert self.extension

    def tearDown(self):
        super(ExtensionTestCaseMixin, self).tearDown()

        if self.extension.info.enabled:
            # Manually shut down the extension first, before we have the
            # extension manager disable it. This will help ensure we have the
            # right state up-front.
            self.extension.shutdown()

            self.extension_mgr.disable_extension(self.extension_class.id)

    def get_extension_manager(self):
        """Return the extension manager used for the tests.

        Subclasses may want to override this to pick a specific extension
        manager, if the project uses more than one. The default behavior is
        to return the first registered extension manager.

        Returns:
            djblets.extensions.manager.ExtensionManager:
            The extension manager used for tests.
        """
        return get_extension_managers()[0]



"""Mixins for integrating token-based authentication into an API."""

from __future__ import unicode_literals

from django.contrib import auth
from django.utils import six

from djblets.webapi.errors import PERMISSION_DENIED


class ResourceAPITokenMixin(object):
    """Augments a WebAPIResource to support API tokens.

    Any WebAPIResource subclass making use of this mixin can accept requests
    backed by an API token, and will restrict the request to that token's
    policy.

    It's recommended that all resources in a project inherit from a base
    resource that inherits both from this mixin and from WebAPIResource. The
    subclass must provide, at a minimum, a value for
    :py:attr:`api_token_model`.
    """

    #: The model class for storing and accessing API token state.
    api_token_model = None

    #: Whether or not a client using API tokens can access this resource.
    api_token_access_allowed = True

    @property
    def policy_id(self):
        """Return the ID used for access policies.

        This defaults to the name of the resource, but can be overridden
        in case the name is not specific enough or there's a conflict.
        """
        return self.name

    def call_method_view(self, request, method, view, *args, **kwargs):
        """Check token access policies and call the API method handler.

        If the client has authenticated with an API token, the token's
        access policies will be checked before invoking the API method
        handler. If the policy disallows this operation, a
        ``PERMISSION_DENIED`` error will be returned.
        """
        # This will associate the token, if any, with the request.
        webapi_token = self.__get_api_token_for_request(request)

        if webapi_token:
            if not self.api_token_access_allowed:
                return PERMISSION_DENIED

            policy = webapi_token.policy
            resources_policy = policy.get('resources')

            if resources_policy:
                resource_id = kwargs.get(self.uri_object_key)

                if not self.is_resource_method_allowed(resources_policy,
                                                       method, resource_id):
                    # The token's policies disallow access to this resource.
                    return PERMISSION_DENIED

        return super(ResourceAPITokenMixin, self).call_method_view(
            request, method, view, *args, **kwargs)

    def is_resource_method_allowed(self, resources_policy, method,
                                   resource_id):
        """Return whether a method can be performed on a resource.

        A method can be performed if a specific per-resource policy allows
        it, and the global policy also allows it.

        The per-resource policy takes precedence over the global policy.
        If, for instance, the global policy blocks and the resource policies
        allows, the method will be allowed.

        If no policies apply to this, then the default is to allow.
        """
        # First check the resource policy. For this, we'll want to look in
        # both the resource ID and the '*' wildcard.
        resource_policy = resources_policy.get(self.policy_id)

        if resource_policy:
            permission = self.__check_resource_policy(
                resource_policy, method, [resource_id, '*'])

            if permission is not None:
                return permission

        # Nothing was found there. Now check in the global policy. Note that
        # there isn't a sub-key of 'resources.*', so we'll check based on
        # resources_policy.
        if '*' in resources_policy:
            permission = self.__check_resource_policy(
                resources_policy, method, ['*'])

            if permission is not None:
                return permission

        return True

    def __check_resource_policy(self, policy, method, keys):
        """Check the policy for a specific resource and method.

        This will grab the resource policy for the given policy ID,
        and see if a given method can be performed on that resource,
        without factoring in any global policy rules.

        If the method is allowed and ``restrict_ids`` is ``True``, this will
        then check if the resource should be blocked based on the ID.

        In case of a conflict, blocked policies always trump allowed
        policies.
        """
        for key in keys:
            sub_policy = policy.get(key)

            if sub_policy:
                # We first want to check the specific values, to see if they've
                # been singled out. If not found, we'll check the wildcards.
                #
                # Blocked values always take precedence over allowed values.
                allowed = sub_policy.get('allow', [])
                blocked = sub_policy.get('block', [])

                if method in blocked:
                    return False
                elif method in allowed:
                    return True
                elif '*' in blocked:
                    return False
                elif '*' in allowed:
                    return True

        return None

    def __get_api_token_for_request(self, request):
        webapi_token = getattr(request, '_webapi_token', None)

        if not webapi_token:
            webapi_token_id = request.session.get('webapi_token_id')

            if webapi_token_id:
                try:
                    webapi_token = self.api_token_model.objects.get(
                        pk=webapi_token_id,
                        user=request.user)
                except self.api_token_model.DoesNotExist:
                    # This token is no longer valid. Log the user out.
                    auth.logout(request)

                request._webapi_token = webapi_token

        return webapi_token

    def _get_queryset(self, request, is_list=False, *args, **kwargs):
        """Return the queryset for the resource.

        This is a specialization of :py:meth:`WebAPIResource._get_queryset()`,
        which imposes further restrictions on the queryset results if using
        a WebAPIToken for authentication that defines a policy.

        Any items in the queryset that are denied by the policy will be
        excluded from the results.
        """
        queryset = super(ResourceAPITokenMixin, self)._get_queryset(
            request, is_list=is_list, *args, **kwargs)

        if is_list:
            # We'll need to filter the list of results down to exclude any
            # that are blocked for GET access by the token policy.
            webapi_token = self.__get_api_token_for_request(request)

            if webapi_token:
                resources_policy = webapi_token.policy.get('resources', {})
                resource_policy = resources_policy.get(self.policy_id)

                if resource_policy:
                    resource_ids = [
                        resource_id
                        for resource_id in six.iterkeys(resource_policy)
                        if (resource_id != '*' and
                            not self.__check_resource_policy(
                                resources_policy, self.policy_id, 'GET',
                                resource_id, True))
                    ]

                    if resource_ids:
                        queryset = queryset.exclude(**{
                            self.model_object_key + '__in': resource_ids,
                        })

        return queryset



"""Base support and implementations for extension hooks.

Extension hooks allow applications to define formal ways to inject logic,
behavior, or more into part of the application.

This module provides the base support for defining an extension hook, through
:py:class:`ExtensionHook` and :py:class:`ExtensionHookPoint`, along with
a utility mixin, :py:class:`AppliesToURLMixin`.

It also provides some built-in hooks for applications and extensions to use:

* :py:class:`DataGridColumnsHook`
* :py:class:`SignalHook`
* :py:class:`TemplateHook`
* :py:class:`URLHook`
"""

from __future__ import unicode_literals

import logging
import uuid
import warnings

from django.template import RequestContext
from django.utils import six

from djblets.deprecation import RemovedInDjblets20Warning
from djblets.util.compat.django.template.loader import render_to_string


logger = logging.getLogger(__name__)


class ExtensionHook(object):
    """The base class for a hook into some part of an application.

    ExtensionHooks are classes that can hook into an
    :py:class:`ExtensionHookPoint` to provide some level of functionality in an
    application. A consuming application should provide a subclass of
    ExtensionHook that will provide functions for getting data or anything else
    that's needed. Extensions may then subclass or initialize that specific
    ExtensionHook.

    A base ExtensionHook subclass must use :py:class:`ExtensionHookPoint`
    as a metaclass. All hooks deriving from that subclass will be registered
    along with that hook point.

    Example:
        .. code-block:: python

           from django.utils import six

           from myproject.nav import register_thing, unregister_thing_id


           @six.add_metaclass(ExtensionHookPoint)
           class ThingHook(ExtensionHook):
               def initialize(self, thing_id):
                   self.thing_id = thing_id
                   register_thing(self.thing_id)

               def shutdown(self):
                   unregister_thing(self.thing_id)

    .. versionchanged:: 1.0

       Starting with Djblets 1.0, extension hooks should implement the
       :py:meth:`initialize` method to handle any initialization. It no longer
       needs to call the parent :py:meth:`shutdown` method, either. However,
       to retain compatibility with older versions, they may still override
       :py:meth:`__init__` and may call the parent :py:meth:`shutdown`. See
       those methods for more details.

    Attributes:
        extension (djblets.extensions.extension.Extension):
            The parent extension, or another object that can act as a
            hook owner.

        hook_state (int):
            The state of the hook. This will be one of
            :py:data:`HOOK_STATE_DISABLED`, :py:data:`HOOK_STATE_ENABLED`,
            :py:data:`HOOK_STATE_DISABLING`, or :py:data:`HOOK_STATE_ENABLING`.
    """

    #: The hook is disabled.
    HOOK_STATE_DISABLED = 0

    #: The hook is enabled.
    HOOK_STATE_ENABLED = 1

    #: The hook is in the process of disabling.
    HOOK_STATE_DISABLING = 2

    #: The hook is in the process of enabling.
    HOOK_STATE_ENABLING = 3

    def __init__(self, extension, *args, **kwargs):
        """Initialize the ExtensionHook.

        This is called when creating an instance of the hook. This will
        call :py:meth:`enable_hook` with the provided arguments, beginning
        the internal initialization process. That will then call
        :py:meth:`initialize`, which is responsible for any initialization
        of state in the subclass.

        Subclasses should override :py:meth:`initialize` in order to provide
        any state initialization, rather than overriding this method.

        .. versionchanged:: 1.0

           Prior to Djblets 1.0, initialization all happened in
           :py:meth:`__init__`. Code that needs to remain compatible with
           older versions should continue to do so, but otherwise this code
           should move to :py:meth:`initialize`.

        Args:
            extension (djblets.extensions.extension.Extension):
                The parent extension, or another object that can act as a
                hook owner.

            start_enabled (bool, optional):
                Whether to enable the hook once constructed. This defaults
                to ``True``.
        """
        self.extension = extension
        self.hook_state = self.HOOK_STATE_DISABLED

        if kwargs.pop('start_enabled', True):
            self.enable_hook(*args, **kwargs)

    @property
    def initialized(self):
        """Whether the hook is initialized and enabled."""
        return self.hook_state == self.HOOK_STATE_ENABLED

    def initialize(self, *args, **kwargs):
        """Initialize the extension hook's state.

        Extension subclasses can perform any custom initialization they need
        here.

        Any additional arguments passed to the hook during construction will be
        passed to this as well.

        While in this function, :py:attr:`hook_state` will be set to
        :py:data:`HOOK_STATE_ENABLING`.

        By default, this does nothing.

        .. versionadded:: 1.0

        Args:
            *args (tuple):
                The list of additional arguments used to initialize the
                hook state.

            **kwargs (dict):
                The additional keyword arguments used to initialize the hook
                state.

        Example:
            .. code-block:: python

               @six.add_metaclass(ExtensionHookPoint)
               class ThingHook(ExtensionHook):
                   def initialize(self, thing_id):
                       self.thing_id = thing_id
                       register_thing(self.thing_id)
        """
        pass

    def shutdown(self):
        """Shut down the extension.

        Extension subclasses can perform any custom cleanup they need here.

        While in this function, :py:attr:`hook_state` will be set to
        :py:data:`HOOK_STATE_DISABLING`.

        .. versionchanged:: 1.0

           This method used to be responsible both for internal cleanup and the
           cleanup of the subclass. Starting in Djblets 1.0, internal cleanup
           has moved to :py:meth:`disable_hook`. Subclasses no longer
           need to call the parent method unless inheriting from a mixin or
           another :py:class:`ExtensionHook` subclass, but should continue to
           do so if they need to retain compatibility with older versions.

        Example:

            .. code-block:: python

               @six.add_metaclass(ExtensionHookPoint)
               class ThingHook(ExtensionHook):
                   def shutdown(self):
                       unregister_thing(self.thing_id)
        """
        if self.hook_state == self.HOOK_STATE_ENABLED:
            # The extension framework will call disable_hook() properly,
            # but a unit test might manually be calling shutdown(). In this
            # case, we want to retain the old behavior, but warn.
            warnings.warn(
                '%(name)s.disable_hook() should be called when manually '
                'disabling the hook instead of %(name)s.shutdown(), starting '
                'in Djblets 1.0.'
                % {
                    'name': self.__class__.__name__,
                },
                RemovedInDjblets20Warning)

            # Don't call shutdown() again, as the subclass might have already
            # dealt with state cleanup.
            self.disable_hook(call_shutdown=False)

    def enable_hook(self, *args, **kwargs):
        """Enable the ExtensionHook, beginning the initialization process.

        This will register the instance of the hook and begin its
        initialization. It takes the same parameters that would be given
        during construction of the hook, allowing disabled hooks to be
        created again with fresh state.

        Subclasses should not override this process. They should instead
        implement :py:meth:`initialize` to handle initialization of the
        state of the hook.

        .. versionadded:: 1.0

        Args:
            *args (tuple):
                The list of additional arguments used to initialize the
                hook state.

            **kwargs (dict):
                The additional keyword arguments used to initialize the hook
                state.
        """
        assert self.hook_state == self.HOOK_STATE_DISABLED

        self.hook_state = self.HOOK_STATE_ENABLING
        self.extension.hooks.add(self)
        self.__class__.add_hook(self)

        self.initialize(*args, **kwargs)
        self.hook_state = self.HOOK_STATE_ENABLED

    def disable_hook(self, call_shutdown=True):
        """Disable the hook, unregistering it from the extension.

        This will unregister the hook and uninitialize it, putting it into
        a disabled state.

        Consumers can call this if they want to turn off hooks temporarily
        without reconstructing the instances later. It's also called
        internally when shutting down an extension.

        .. versionadded:: 1.0

        Args:
            call_shutdown (bool, optional):
                Whether to call :py:meth:`shutdown`. This should always be
                ``True`` unless called internally.
        """
        assert self.hook_state == self.HOOK_STATE_ENABLED

        self.hook_state = self.HOOK_STATE_DISABLING

        if call_shutdown:
            self.shutdown()

        self.__class__.remove_hook(self)
        self.hook_state = self.HOOK_STATE_DISABLED


class ExtensionHookPoint(type):
    """A metaclass used for base Extension Hooks.

    Base :py:class:`ExtensionHook` classes use :py:class:`ExtensionHookPoint`
    as a metaclass. This metaclass stores the list of registered hooks that
    an :py:class:`ExtensionHook` will automatically register with.
    """
    def __init__(cls, name, bases, attrs):
        super(ExtensionHookPoint, cls).__init__(name, bases, attrs)

        if not hasattr(cls, "hooks"):
            cls.hooks = []

    def add_hook(cls, hook):
        """Adds an ExtensionHook to the list of active hooks.

        This is called automatically by :py:class:`ExtensionHook`.
        """
        cls.hooks.append(hook)

    def remove_hook(cls, hook):
        """Removes an ExtensionHook from the list of active hooks.

        This is called automatically by :py:class:`ExtensionHook`.
        """
        cls.hooks.remove(hook)


class AppliesToURLMixin(object):
    """A mixin for hooks to allow restricting to certain URLs.

    This provides an :py:meth:`applies_to` function for the hook that can be
    used by consumers to determine if the hook should apply to the current
    page.
    """

    def initialize(self, apply_to=[], *args, **kwargs):
        """Initialize the mixin for a hook.

        Args:
            apply_to (list, optional):
                A list of URL names that the hook will apply to by default.
        """
        self.apply_to = apply_to

        super(AppliesToURLMixin, self).initialize(*args, **kwargs)

    def applies_to(self, request):
        """Returns whether or not this hook applies to the page.

        This will determine whether any of the URL names provided in
        ``apply_to`` matches the current requested page.
        """
        return (not self.apply_to or
                (request.resolver_match and
                 request.resolver_match.url_name in self.apply_to))


@six.add_metaclass(ExtensionHookPoint)
class DataGridColumnsHook(ExtensionHook):
    """Adds columns to a datagrid.

    This hook allows an extension to register new columns to any datagrid.
    These columns can be added by the user, rearranged, and sorted, like
    any other column.

    Each column must have an id already set, and it must be unique.
    """

    def initialize(self, datagrid_cls, columns):
        """Initialize the hook.

        Args:
            datagrid_cls (type):
                The specific datagrid class that will include this
                registered list of columns as possible options.

            columns (list):
                A list of :py:class:`~djblets.datagrid.grids.Column`
                instances to register on the datagrid.
        """
        self.datagrid_cls = datagrid_cls
        self.columns = columns

        for column in columns:
            self.datagrid_cls.add_column(column)

    def shutdown(self):
        for column in self.columns:
            self.datagrid_cls.remove_column(column)


@six.add_metaclass(ExtensionHookPoint)
class URLHook(ExtensionHook):
    """Custom URL hook.

    A hook that installs custom URLs. These URLs reside in a project-specified
    parent URL.
    """

    def initialize(self, patterns):
        """Initialize the hook.

        Args:
            patterns (list):
                The list of :py:func:`~django.conf.urls.url` entries
                comprising the URLs to register.
        """
        self.patterns = patterns
        self.dynamic_urls = self.extension.extension_manager.dynamic_urls
        self.dynamic_urls.add_patterns(patterns)

    def shutdown(self):
        self.dynamic_urls.remove_patterns(self.patterns)


@six.add_metaclass(ExtensionHookPoint)
class SignalHook(ExtensionHook):
    """Connects to a Django signal.

    This will handle connecting to a signal, calling the specified callback
    when fired. It will disconnect from the signal when the extension is
    disabled.

    The callback will also be passed an extension= keyword argument pointing
    to the extension instance.
    """

    def initialize(self, signal, callback, sender=None, sandbox_errors=True):
        """Initialize the hook.

        Args:
            signal (django.dispatch.Signal):
                The signal to connect to.

            callback (callable):
                The function to call when the signal is fired.

            sender (object or class, optional):
                The sender argument to pass to the signal connection.
                See :py:meth:`~django.core.dispatch.Signal.send` for more
                information.

            sandbox_errors (bool, optional):
                If ``True``, errors coming from ``callback`` will be
                sandboxed, preventing them from reaching the code that
                fired the signal. The error will instead be logged and
                then ignored.
        """
        self.signal = signal
        self.callback = callback
        self.dispatch_uid = uuid.uuid1()
        self.sender = sender
        self.sandbox_errors = sandbox_errors

        signal.connect(self._wrap_callback, sender=self.sender, weak=False,
                       dispatch_uid=self.dispatch_uid)

    def shutdown(self):
        self.signal.disconnect(dispatch_uid=self.dispatch_uid,
                               sender=self.sender)

    def _wrap_callback(self, **kwargs):
        """Wraps a callback function, passing extra parameters and sandboxing.

        This will call the callback with an extension= keyword argument,
        and sandbox any errors (if sandbox_errors is True).
        """
        try:
            self.callback(extension=self.extension, **kwargs)
        except Exception as e:
            logger.exception('Error when calling %r from SignalHook: %s',
                             self.callback, e)

            if not self.sandbox_errors:
                raise


@six.add_metaclass(ExtensionHookPoint)
class TemplateHook(AppliesToURLMixin, ExtensionHook):
    """Custom templates hook.

    A hook that renders a template at hook points defined in another template.
    """

    _by_name = {}

    def initialize(self, name, template_name=None, apply_to=[],
                   extra_context={}):
        """Initialize the hook.

        Args:
            name (unicode):
                The name of the template hook point that should render this
                template. This is application-specific.

            template_name (unicode, optional):
                The name of the template to render.

            apply_to (list, optional):
                The list of URL names where this template should render.
                By default, all templates containing the template hook
                point will render this template.

            extra_context (dict):
                Extra context to include when rendering the template.
        """
        super(TemplateHook, self).initialize(apply_to=apply_to)

        self.name = name
        self.template_name = template_name
        self.extra_context = extra_context

        if name not in self.__class__._by_name:
            self.__class__._by_name[name] = [self]
        else:
            self.__class__._by_name[name].append(self)

    def shutdown(self):
        self.__class__._by_name[self.name].remove(self)

    def render_to_string(self, request, context):
        """Renders the content for the hook.

        By default, this renders the provided template name to a string
        and returns it.
        """
        context_data = {
            'extension': self.extension,
        }
        context_data.update(self.get_extra_context(request, context))
        context_data.update(self.extra_context)

        # Note that context.update implies a push().
        context.update(context_data)

        s = render_to_string(template_name=self.template_name,
                             context=context,
                             request=request)

        context.pop()

        return s

    def get_extra_context(self, request, context):
        """Returns extra context for the hook.

        Subclasses can override this to provide additional context
        dynamically beyond what's passed in to the constructor.

        By default, an empty dictionary is returned.
        """
        return {}

    @classmethod
    def by_name(cls, name):
        return cls._by_name.get(name, [])


class BaseRegistryHook(ExtensionHook):
    """A hook for registering an item with a registry.

    This hook should not be used directly. Instead, it should be subclassed
    with the :py:attr:`registry` attribute set.

    Subclasses must use the :py:class:`ExtensionHookPoint` metaclass.
    """

    #: The registry to register items with.
    registry = None

    def initialize(self, item):
        """Initialize the registry hook with the item.

        Args:
            item (object):
                The object to register.
        """
        self.item = item
        self.registry.register(item)

    def shutdown(self):
        """Shut down the registry hook and unregister the item."""
        self.registry.unregister(self.item)


class BaseRegistryMultiItemHook(ExtensionHook):
    """A hook for registering multiple items with a registry.

    This hook should not be used directly. Instead, it should be subclassed
    with the :py:attr:`registry` attribute set.

    Subclasses must use the :py:class:`ExtensionHookPoint` metaclass.
    """

    #: The registry to register items with.
    registry = None

    def initialize(self, items):
        """Initialize the registry hook with the list of items.

        Args:
            items (list):
                The list of items to register.
        """
        self.items = items

        registered_items = []

        for item in items:
            try:
                self.registry.register(item)
            except Exception:
                # If there's an error, first unregister all existing items and
                # then re-raise the error.
                for item in registered_items:
                    try:
                        self.registry.unregister(item)
                    except:
                        pass

                raise

            registered_items.append(item)

    def shutdown(self):
        """Shut down the registry hook and unregister the items."""
        for item in self.items:
            self.registry.unregister(item)



"""Serial numbers used for front-end and back-end caching keys/URLs.

These functions generate stable serial numbers based on the content shipped
with the product (static media files, templates, locales). These can be used
in various cache keys, ETags, and URLs to help keep content in cache until
it changes on disk.
"""

from __future__ import unicode_literals

import importlib
import logging
import os

from django.conf import settings


logger = logging.getLogger(__name__)


def generate_media_serial():
    """Generate a media serial number for static media files.

    The media serial number can be appended to a media filename in order to
    make a URL that can be cached forever without fear of change.  The next
    time the file is updated and the server is restarted, a new path will be
    accessed and cached.

    This will crawl the media files (using directories in
    :setting:`MEDIA_SERIAL_DIRS` if specified, or all of
    :django:setting:`STATIC_ROOT` otherwise), figuring out the latest
    timestamp, and return that value.
    """
    MEDIA_SERIAL = getattr(settings, "MEDIA_SERIAL", 0)

    if not MEDIA_SERIAL:
        media_dirs = getattr(settings, "MEDIA_SERIAL_DIRS", ["."])

        for media_dir in media_dirs:
            media_path = os.path.join(settings.STATIC_ROOT, media_dir)

            for root, dirs, files in os.walk(media_path):
                for name in files:
                    mtime = int(os.stat(os.path.join(root, name)).st_mtime)

                    if mtime > MEDIA_SERIAL:
                        MEDIA_SERIAL = mtime

        setattr(settings, "MEDIA_SERIAL", MEDIA_SERIAL)


def generate_ajax_serial():
    """Generate a template-based AJAX serial number for requests and ETags.

    The serial number can be appended to filenames involving dynamic loads of
    URLs in order to make a URL that can be cached forever without fear of
    change.

    This will crawl the template files (using directories in
    :django:setting:`TEMPLATE_DIRS`), figuring out the latest timestamp, and
    return that value.
    """
    AJAX_SERIAL = getattr(settings, "AJAX_SERIAL", 0)

    if not AJAX_SERIAL:
        template_dirs = getattr(settings, "TEMPLATE_DIRS", ["."])

        for template_path in template_dirs:
            for root, dirs, files in os.walk(template_path):
                for name in files:
                    mtime = int(os.stat(os.path.join(root, name)).st_mtime)

                    if mtime > AJAX_SERIAL:
                        AJAX_SERIAL = mtime

        setattr(settings, "AJAX_SERIAL", AJAX_SERIAL)


def generate_locale_serial(packages):
    """Generate a locale serial for the given set of packages.

    This will be equal to the most recent mtime of all the .mo files that
    contribute to the localization of the given packages.

    Unlike the other serial-generation functions, this will return the
    value, rather than setting it on ``settings``.

    Args:
        packages (list of unicode):
            A list of Python module paths containing :file:`locale`
            directories.

    Returns:
        int: The resulting serial number.
    """
    serial = 0

    paths = []
    for package in packages:
        try:
            p = importlib.import_module(package)
            path = os.path.join(os.path.dirname(p.__file__), 'locale')
            paths.append(path)
        except Exception as e:
            logger.exception(
                'Failed to import package %s to compute locale serial: %s',
                package, e)

    for locale_path in paths:
        for root, dirs, files in os.walk(locale_path):
            for name in files:
                if name.endswith('.mo'):
                    mtime = int(os.stat(os.path.join(root, name)).st_mtime)
                    if mtime > serial:
                        serial = mtime

    return serial


def generate_cache_serials():
    """Generate both static media and AJAX serial numbers.

    This is a wrapper around :py:func:`generate_media_serial` and
    :py:func:`generate_ajax_serial`, which generates all the serial numbers
    in one go.

    This should be called early in the startup, such as in the site's
    main :file:`urls.py`.
    """
    generate_media_serial()
    generate_ajax_serial()



#
# core.py -- Core classes for webapi
#
# Copyright (c) 2007-2009  Christian Hammond
# Copyright (c) 2007-2009  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import unicode_literals

import warnings

from djblets.deprecation import RemovedInDjblets20Warning
from djblets.webapi.decorators import SPECIAL_PARAMS
from djblets.webapi.encoders import (BasicAPIEncoder, JSONEncoderAdapter,
                                     WebAPIEncoder, XMLEncoderAdapter,
                                     get_registered_encoders)
from djblets.webapi.responses import (WebAPIResponse, WebAPIResponseError,
                                      WebAPIResponseFormError,
                                      WebAPIResponsePaginated)


warnings.warn('djblets.webapi.core is deprecated', RemovedInDjblets20Warning)


__all__ = [
    'BasicAPIEncoder',
    'JSONEncoderAdapter',
    'SPECIAL_PARAMS',
    'WebAPIEncoder',
    'WebAPIResponse',
    'WebAPIResponseError',
    'WebAPIResponseFormError',
    'WebAPIResponsePaginated',
    'XMLEncoderAdapter',
    'get_registered_encoders',
]



"""Consent tracking and storage capabilities."""

from __future__ import unicode_literals

import hashlib
from importlib import import_module

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.utils import six

from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.db.query import get_object_or_none
from djblets.privacy.consent.base import Consent
from djblets.privacy.consent.registry import get_consent_requirements_registry
from djblets.privacy.models import StoredConsentData


class BaseConsentTracker(object):
    """Base class for a consent tracker.

    Consent trackers are responsible for storing and looking up consent
    information from a source. This can be the database, an external record
    kept outside the database, a secure vault, or some other location as
    supported by the specific consent tracker being used.

    By default, :py:class:`DatabaseConsentTracker` is used. This can be changed
    through ``settings.DJBLETS_PRIVACY_CONSENT_TRACKER``.
    """

    def get_audit_identifier(self, user):
        """Return an identifier to store for audit purposes.

        This is used to help locate stored consent, perhaps after a user
        account was unregistered, in order to prove later that a user did
        or did not consent to something.

        Data used in this identifier should never be used for any purpose
        other than locating the record for legal reasons. Stored audit
        identifiers may be erased at any time.

        By default, this returns a SHA256 hash representing the user's e-mail
        address, allowing lookups for known e-mail addresses without risking
        exposure. Subclasses can modify this.

        Args:
            user (django.contrib.auth.models.User):
                The user to return an identifier for.

        Returns:
            str:
            An identifier to store to help locate the record of consent.
        """
        return hashlib.sha256(user.email.encode('utf-8')).hexdigest()

    def record_consent_data(self, user, consent_data):
        """Record information on a consent decision made by a user.

        Args:
            user (django.contrib.auth.models.User):
                The user to record the consent decision for.

            consent_data (djblets.privacy.consent.base.ConsentData):
                Data on the consent decision.
        """
        self.record_consent_data_list(user, [consent_data])

    def get_consent(self, user, requirement_id):
        """Return the user's consent decision for a given requirement.

        Args:
            user (django.contrib.auth.models.User):
                The user who made a decision on consent for a requirement.

            requirement_id (unicode):
                The ID of the requirement to check on consent for.

        Returns:
            djblets.privacy.consent.base.Consent:
            The user's consent decision for the requirement.
        """
        consents = self.get_all_consent(user) or {}

        return consents.get(requirement_id, Consent.UNSET)

    def get_pending_consent_requirements(self, user):
        """Return a list of consent requirements that are pending decisions.

        This can be used to determine whether a user needs to be immediately
        shown a UI for deciding on consent.

        Args:
            user (django.contrib.auth.models.User):
                The user to check consent decisions for.

        Returns:
            list of djblets.privacy.consent.base.ConsentRequirement:
            The list of consent requirements pending decisions.
        """
        all_consent = self.get_all_consent(user)

        return [
            consent_requirement
            for consent_requirement in get_consent_requirements_registry()
            if consent_requirement.requirement_id not in all_consent
        ]

    def record_consent_data_list(self, user, consent_data_list):
        """Record a list of all consent data made by a user.

        Args:
            user (django.contrib.auth.models.User):
                The user to record the consent data for.

            consent_data_list (list of
                               djblets.privacy.consent.base.ConsentData):
                A list of consent data to record.
        """
        self.store_recorded_consent_data_list(user, consent_data_list)
        cache.delete(make_cache_key(self._get_user_cache_key(user)))

    def get_all_consent(self, user):
        """Return all consent decisions made by a given user.

        It's important to note that a user may not have made a decision on
        consent for a given registered requirement, in which case the results
        will not include an entry for that requirement.

        Args:
            user (django.contrib.auth.models.User):
                The user to return all consent information for.

        Returns:
            dict:
            A dictionary of
            :py:class:`~djblets.privacy.consent.base.BaseConsentRequirement`
            IDs to :py:class:`~djblets.privacy.consent.base.Consent` values.
        """
        return cache_memoize(self._get_user_cache_key(user),
                             lambda: self.get_all_consent_uncached(user))

    def store_recorded_consent_data_list(self, user, consent_data_list):
        """Record a list of all consent data made by a user.

        Args:
            user (django.contrib.auth.models.User):
                The user to record the consent data for.

            consent_data_list (list of
                               djblets.privacy.consent.base.ConsentData):
                A list of consent data to record.
        """
        raise NotImplementedError

    def get_all_consent_uncached(self, user):
        """Return all consent decisions made by a given user from the backend.

        This is used by :py:meth:`get_all_consent` to return data directly
        from the backend, bypassing the cache.

        Args:
            user (django.contrib.auth.models.User):
                The user to return all consent information for.

        Returns:
            dict:
            A dictionary of
            :py:class:`~djblets.privacy.consent.base.BaseConsentRequirement`
            IDs to :py:class:`~djblets.privacy.consent.base.Consent` values.
        """
        raise NotImplementedError

    def _get_user_cache_key(self, user):
        """Return a consent cache key for the provided user.

        Args:
            user (django.contrib.auth.models.User):
                The user to generate the cache key for.

        Returns:
            unicode:
            The resulting cache key.
        """
        return 'privacy-consent:%s' % user.pk


class DatabaseConsentTracker(BaseConsentTracker):
    """A consent tracker that stores results in the database.

    This consent tracker keeps the current consent information and audit
    history in the database for fast and easy lookup. This is the default
    consent tracker used.
    """

    #: The model used to store consent data in the database.
    model = StoredConsentData

    def store_recorded_consent_data_list(self, user, consent_data_list):
        """Store a recorded list of all consent data made by a user.

        Args:
            user (django.contrib.auth.models.User):
                The user to record the consent data for.

            consent_data_list (list of
                               djblets.privacy.consent.base.ConsentData):
                A list of consent data to record.
        """
        stored_consent = self.model.objects.get_or_create(user=user)[0]

        # Update this in case it has changed since the last recording.
        audit_identifier = self.get_audit_identifier(user)
        stored_consent.audit_identifier = audit_identifier

        # These just cut down on attribute access and verbosity.
        grants = stored_consent.consent_grants
        audit_trail = stored_consent.audit_trail

        for consent_data in consent_data_list:
            requirement_id = consent_data.requirement_id
            audit_info = consent_data.serialize_audit_info(audit_identifier)

            grants[requirement_id] = consent_data.granted
            audit_trail.setdefault(requirement_id, []).insert(0, audit_info)

        stored_consent.save()

        user._djblets_stored_consent = stored_consent

    def get_all_consent_uncached(self, user):
        """Return all consent decisions made by a given user.

        It's important to note that a user may not have made a decision on
        consent for a given registered requirement, in which case the results
        will not include an entry for that requirement.

        Args:
            user (django.contrib.auth.models.User):
                The user to return all consent information for.

        Returns:
            dict:
            A dictionary of
            :py:class:`~djblets.privacy.consent.base.BaseConsentRequirement`
            IDs to :py:class:`~djblets.privacy.consent.base.Consent` values.
        """
        stored_consent = self._get_stored_consent_for_user(user)
        result = {}

        if stored_consent:
            for key, value in six.iteritems(stored_consent.consent_grants):
                if value:
                    result[key] = Consent.GRANTED
                else:
                    result[key] = Consent.DENIED

        return result

    def _get_stored_consent_for_user(self, user):
        """Return stored consent data for a user.

        This will look up the consent data and cache it on the user, speeding
        up subsequent calls.

        Args:
            user (django.contrib.auth.models.User):
                The user to retrieve and cache stored consent data for.

        Returns:
            djblets.privacy.models.StoredConsentData:
            The stored consent data for the user, or ``None`` if there's no
            existing data on record.
        """
        if not hasattr(user, '_djblets_stored_consent'):
            user._djblets_stored_consent = \
                get_object_or_none(self.model, user=user)

        return user._djblets_stored_consent


_consent_tracker = None


def get_consent_tracker():
    """Return the registered consent tracker instance.

    Returns:
        BaseConsentTracker:
        The consent tracker instance.
    """
    global _consent_tracker

    if _consent_tracker is None:
        tracker_class_name = getattr(
            settings,
            'DJBLETS_PRIVACY_CONSENT_TRACKER',
            '%s.%s' % (DatabaseConsentTracker.__module__,
                       DatabaseConsentTracker.__name__))

        module_name, class_name = tracker_class_name.rsplit('.', 1)

        try:
            _consent_tracker = getattr(import_module(module_name),
                                       class_name)()
        except (AttributeError, ImportError):
            raise ImproperlyConfigured(
                'settings.DJBLETS_CONSENT_TRACKER must be set to a valid '
                'class path.')

    return _consent_tracker


def clear_consent_tracker():
    """Clear the registered consent tracker instance.

    This is primarily useful for unit tests. It will remove the consent
    tracker, causing it to be re-instantiated based on settings on the next
    attempt.
    """
    global _consent_tracker

    _consent_tracker = None



from __future__ import unicode_literals

from django.http import HttpResponse
from django.shortcuts import render
from django.template.context import RequestContext
from django.utils.six.moves import http_client
from django.utils.six.moves.urllib.error import URLError
from django.utils.six.moves.urllib.request import urlopen

from djblets.cache.backend import cache_memoize
from djblets.util.compat.django.template.loader import render_to_string


DEFAULT_EXPIRATION = 2 * 24 * 60 * 60  # 2 days


def view_feed(request, url, template_name="feedview/feed-page.html",
              cache_expiration=DEFAULT_EXPIRATION, extra_context={}):
    """
    Renders an RSS or Atom feed using the given template. This will use
    a cached copy if available in order to reduce hits to the server.
    """
    def fetch_feed():
        import feedparser

        data = urlopen(url).read()

        parser = feedparser.parse(data)

        context = {
            'parser': parser,
        }
        context.update(extra_context)

        return render_to_string(template_name, context, request)

    try:
        return HttpResponse(cache_memoize("feed-%s" % url, fetch_feed,
                            cache_expiration,
                            force_overwrite=('reload' in request.GET)))
    except (URLError, http_client.HTTPException) as e:
        context = {
            'error': e,
        }
        context.update(extra_context)

        return render(request, template_name, context)



"""Utility functions for working with template contexts."""

from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured

try:
    # Django >= 1.8
    from django.template import engines as template_engines

    get_standard_processors = None
except ImportError:
    # Django <= 1.7
    from django.template.context import get_standard_processors

    template_engines = None


def get_default_template_context_processors(engine_name='django'):
    """Return all default template context processors for an engine.

    Args:
        engine_name (unicode):
            When run on Django 1.8 or higher, this specifies which template
            engine's context processors to return. On Django 1.7 and older,
            this is ignored.

    Returns:
        list of callable:
        The list of template context processors.

    Raises:
        django.core.exceptions.ImproperlyConfigured:
            The specified template engine wasn't valid on this version of
            Django.
    """
    if template_engines is None:
        return get_standard_processors()
    else:
        try:
            template_engine = template_engines[engine_name]
        except KeyError:
            raise ImproperlyConfigured(
                'The "%s" template engine must be defined.'
                % engine_name)

        return template_engine.engine.template_context_processors
